Type: Package
Package: ONAM
Title: Fitting Interpretable Neural Additive Models Using
        Orthogonalization
Version: 1.0.1
Authors@R: 
    person("David", "Köhler", , "koehler@imbie.uni-bonn.de", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0006-0027-4046"))
Description: An algorithm for fitting interpretable additive neural
    networks for identifiable and visualizable feature effects using post
    hoc orthogonalization. Fit custom neural networks intuitively using
    established 'R' 'formula' notation, including interaction effects of
    arbitrary order while preserving identifiability to enable a
    functional decomposition of the prediction function. For more details see 
    Koehler et al. (2025) <doi:10.1038/s44387-025-00033-7>.
License: MIT + file LICENSE
BugReports: https://github.com/Koehlibert/ONAM_R/issues
Depends: keras3, reticulate
Imports: dplyr, scales, rlang, ggplot2, pROC
Suggests: akima, RColorBrewer, testthat (>= 3.0.0)
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-01-26 11:15:34 UTC; koehler
Author: David Köhler [aut, cre] (ORCID:
    <https://orcid.org/0009-0006-0027-4046>)
Maintainer: David Köhler <koehler@imbie.uni-bonn.de>
Repository: CRAN
Date/Publication: 2026-01-26 11:50:02 UTC
Built: R 4.6.0; ; 2026-04-29 05:52:03 UTC; unix

Type: Package
Package: ONAM
Title: Fitting Interpretable Neural Additive Models Using
        Orthogonalization
Version: 1.1.0
Authors@R: 
    person("David", "Köhler", , "koehler@imbie.uni-bonn.de", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0006-0027-4046"))
Description: An algorithm for fitting interpretable additive neural
    networks for identifiable and visualizable feature effects using post
    hoc orthogonalization. Fit custom neural networks intuitively using
    established 'R' 'formula' notation, including interaction effects of
    arbitrary order while preserving identifiability to enable a
    functional decomposition of the prediction function. For more details see 
    Koehler et al. (2025) <doi:10.1038/s44387-025-00033-7>.
License: MIT + file LICENSE
BugReports: https://github.com/Koehlibert/ONAM_R/issues
Depends: keras3, reticulate
Imports: dplyr, scales, rlang, ggplot2, pROC
Suggests: akima, RColorBrewer, testthat (>= 3.0.0)
Encoding: UTF-8
Config/testthat/edition: 3
NeedsCompilation: no
Config/roxygen2/version: 8.0.0
Packaged: 2026-08-24 08:26:46 UTC; koehler
Author: David Köhler [aut, cre] (ORCID:
    <https://orcid.org/0009-0006-0027-4046>)
Maintainer: David Köhler <koehler@imbie.uni-bonn.de>
Repository: CRAN
Date/Publication: 2026-08-24 09:20:15 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,gen_sobol)
S3method(plot,var_decomp)
S3method(predict,onam)
S3method(print,gen_sobol)
S3method(print,summary.onam)
S3method(print,var_decomp)
S3method(summary,onam)
export(build_dnn)
export(decompose)
export(gen_sobol)
export(install_conda_env)
export(load_onam)
export(onam)
export(plot_inter_effect)
export(plot_main_effect)
export(save_onam)
importFrom(dplyr,"%>%")
importFrom(keras3,fit)
importFrom(reticulate,py_available)
importFrom(rlang,.data)
importFrom(stats,predict)

# Orthogonal neural additive models for interpretable machine learning by functional decomposition of black-box models into explainable predictor effects

## Install package:

If this is the first time using keras or tensorflow in R, you need to run `keras3::install_keras()`.
For certain systems, especially windows systems, `install_keras` may not succeed. In this case, use 
`install_conda_env`
to create a conda environment and install relevant python modules.

This readme will guide you through the workflow of specifying, fitting and analysing an orthogonal neural additive model.


## Simulate example data

``` r
# Create training data
n <- 1000
x1 <- runif(n, -2, 2)
x2 <- runif(n, -2, 2)
x3 <- runif(n, -2, 2)
noise <- rnorm(n, 0, 1)
y <- sin(x1) + dt(x2, 1) * 4 + 2 * x3 +
  x1 * x2 + noise
train <- cbind(x1, x2, x3, y)
```

## Specify model architecture(s)
Different feature (interaction) effects can have different model architectures, varying in type and complexity. These architectures are to be specified in the `list_of_deep_models`-argument.

``` r
simple_model <- function(inputs) {
  outputs <- inputs %>%
    layer_dense(units = 128, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 64, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 32, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 16, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 8, activation = "relu", use_bias = TRUE) %>%
    keras::layer_dense(units = 1, activation = "linear", use_bias = TRUE)
  keras_model(inputs, outputs)
}
complex_model <- function(inputs) {
  outputs <- inputs %>%
    layer_dense(units = 512, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 256, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 128, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 64, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 32, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 16, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 8, activation = "relu", use_bias = TRUE) %>%
    layer_dense(units = 1, activation = "linear", use_bias = TRUE)
    keras_model(inputs, outputs)
}

list_of_deep_models = list(simple = simple_model,
                           complex = complex_model)
```

## Specify effects of interest
The effects to be fitted are supplied in a `formula`-object. The names of the functions for each feature have to correspond to the names in the `list_of_deep_models`-argument.

``` r
model_formula = y ~ simple(x1) + simple(x2) + simple(x3) + 
  complex(x1, x2) + complex(x1, x2, x3)
```

## Fit onam model

An onam model can then be fitted according to the model formula. An ensembling strategy is used, here with 2 ensemble members. Further model fitting parameters can be specified, such as verbosity, callbacks or number of training steps.

``` r
mod <- onam(formula = model_formula, 
            list_of_deep_models = list_of_deep_models,
            data = train, n_ensemble = 2, epochs = 100)
```

## Model evaluation

The fitted model can be investigated regarding goodness of fit and degree of interpretability through the `summary()`-function.

``` r
summary(mod)
```
Fitted feature effects can be visualized in `ggplot`-figures:

``` r
plot_main_effect(mod, "x1")
plot_inter_effect(mod, "x1", "x2", interpolate = TRUE)
```

#' Set up conda environment for keras functionality
#'
#' @description
#' Helper function to install Keras and packages necessary for package
#' functionality into a conda environment. Use this function if
#' [keras3::install_keras()] does not work, esp. on windows machines.
#'
#'
#' @param envname Name for the conda environment to be created.
#' @param python_version Python version to be installed in the conda environment.
#' @param overwrite Should an existing conda environment of name `envname` be
#' overwritten if present?
#'
#' @return No return value, called for side effects
#' @seealso [keras3::install_keras()]
#' @export
install_conda_env <- function(
    envname = "r-keras",
    python_version = "python=3.10",
    overwrite = FALSE
) {
  conda_bin <- reticulate::conda_binary()
  envs <- reticulate::conda_list(conda = conda_bin)$name

  if (envname %in% envs) {

    if (!overwrite) {
      if (interactive()) {
        ans <- readline(
          paste0(
            "Conda environment '", envname,
            "' already exists. Overwrite it? [y/N]: "
          )
        )
        overwrite <- tolower(ans) %in% c("y", "yes")
      }
    }

    if (!overwrite) {
      message("Using existing conda environment: ", envname)
      reticulate::use_condaenv(envname, required = TRUE)
      return(invisible(FALSE))
    }

    message("Removing existing conda environment: ", envname)
    reticulate::conda_remove(envname, conda = conda_bin)
  }

  message("Creating conda environment: ", envname)
  reticulate::conda_create(
    envname,
    packages = python_version,
    conda = conda_bin
  )

  reticulate::conda_install(
    envname,
    packages = c("tensorflow", "keras"),
    conda = conda_bin
  )

  reticulate::use_condaenv(envname, required = TRUE)

  reticulate::py_run_string(
    "import sys, tensorflow as tf, keras; print(sys.version); print(tf.__version__); print(keras.__version__)"
  )

  invisible(TRUE)
}

#' Save a fitted onam model to disk
#'
#' @description
#' An `onam` object holds one or more fitted 'keras' submodels per ensemble
#' member (in `object$ensemble[[i]]$model_list`), which cannot be
#' serialized with `saveRDS()` alone. This function saves each submodel to
#' its own `.keras` file and the remaining orthogonalization metadata
#' (weights, `model_info`, data, predictions, ...) to a single `.rds` file,
#' all inside `dir`. Use [load_onam] to restore the object.
#'
#' @param object Object of class `onam` as returned by [onam].
#' @param dir Directory the model is saved to. Created if it does not exist.
#' @param overwrite Should an existing directory of the same name be
#' overwritten?
#' @returns No return value, called for side effects.
#' @seealso [load_onam]
#' @export
save_onam <- function(object, dir, overwrite = FALSE) {
  stopifnot(inherits(object, "onam"))
  if (dir.exists(dir)) {
    if (!overwrite) {
      stop("`dir` already exists. Set `overwrite = TRUE` to overwrite.",
           call. = FALSE)
    }
  } else {
    dir.create(dir, recursive = TRUE)
  }
  submodel_dir <- file.path(dir, "submodels")
  dir.create(submodel_dir, showWarnings = FALSE)
  n_ensemble <- length(object$ensemble)
  for (i in seq_len(n_ensemble)) {
    orders <- names(object$ensemble[[i]]$model_list)
    for (order in orders) {
      model_indices <- seq_along(object$ensemble[[i]]$model_list[[order]])
      for (model_idx in model_indices)
      {
        keras3::save_model(
          object$ensemble[[i]]$model_list[[order]][[model_idx]],
          file.path(
            submodel_dir,
            paste0("ensemble", i, "_", order, "_", model_idx, ".keras")
          ),
          overwrite = TRUE
        )
      }
    }
    # Keras models can't round-trip through saveRDS; keep the order names
    # and how many models each order holds so load_onam() knows which
    # files to reload and where they go.
    model_counts <- vapply(orders, function(order) {
      length(object$ensemble[[i]]$model_list[[order]])
    }, integer(1))
    names(model_counts) <- orders
    object$ensemble[[i]]$model_list <- model_counts
  }
  saveRDS(object, file.path(dir, "onam_meta.rds"))
  invisible(NULL)
}
#' Load a fitted onam model from disk
#'
#' @param dir Directory as created by [save_onam].
#' @returns Returns the restored object of class `onam`.
#' @seealso [save_onam]
#' @export
load_onam <- function(dir) {
  if (!require_keras()) {
    invisible(return(NULL))
  }
  meta_path <- file.path(dir, "onam_meta.rds")
  if (!file.exists(meta_path)) {
    stop("No onam model found in `dir`.", call. = FALSE)
  }
  object <- readRDS(meta_path)
  submodel_dir <- file.path(dir, "submodels")
  n_ensemble <- length(object$ensemble)
  for (i in seq_len(n_ensemble)) {
    model_counts <- object$ensemble[[i]]$model_list
    orders <- names(model_counts)
    model_list <- lapply(orders, function(order) {
      lapply(seq_len(model_counts[[order]]), function(model_idx) {
        keras3::load_model(file.path(
          submodel_dir,
          paste0("ensemble", i, "_", order, "_", model_idx, ".keras")
        ))
      })
    })
    names(model_list) <- orders
    object$ensemble[[i]]$model_list <- model_list
  }
  object
}

test_that("Saving works", {
  skip_if_no_conda_env()
  tmp_dir <- tempfile("onam_persistence_")
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  x4 <- runif(n, -2, 2)
  y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3), pweibull(-x2, shape = 0.5)) +
    dt(x3, 1) * 4 +
    x1 * x2
  data_train <- cbind(x1, x2, x3, x4, y)
  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3)
  mod1 <- build_dnn(c(16,8,1))
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  mod <- onam(
    f1,
    list_of_deep_models,
    data_train,
    n_ensemble = 1,
    epochs = 10,
    progresstext = FALSE,
    verbose = 0
  )
  save_onam(mod, tmp_dir)
  mod_loaded <- load_onam(tmp_dir)
  out <- predict(mod, data_train)
  out_loaded <- predict(mod_loaded)
  expect_equal(out$predictions, out_loaded$predictions)
  expect_equal(out$feature_effects, out_loaded$feature_effects)
})

#' Get variance decomposition of orthogonal neural additive model
#' @param object Either model of class `onam` as returned from [onam] or
#' model evaluation outcome as returned from [predict.onam]
#' @param data Data for which the model is to be evaluated. If \code{NULL}
#' (DEFAULT), the data from model fitting is used.
#' @returns Returns a named vector of percentage of variance explained by each
#' interaction order.
#' @examplesIf reticulate::py_module_available(tensorflow)
#' \donttest{
#' # Basic example for a simple ONAM-model
#' # Create training data
#' n <- 1000
#' x1 <- runif(n, -2, 2)
#' x2 <- runif(n, -2, 2)
#' y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
#'   pweibull(-x2, shape = 0.5)) +
#'   x1 * x2
#' data_train <- cbind(x1, x2, y)
#' # Define model
#' model_formula <- y ~ mod1(x1) + mod1(x2) +
#'   mod1(x1, x2)
#' mod1 <- function(inputs) {
#'   outputs <- inputs %>%
#'     layer_dense(units = 16, activation = "relu") %>%
#'     layer_dense(units = 8, activation = "linear",
#'                 use_bias = TRUE) %>%
#'     layer_dense(units = 1, activation = "linear",
#'                 use_bias = TRUE)
#'   keras_model(inputs, outputs)
#' }
#' list_of_deep_models <- list(mod1 = mod1)
#' # Fit model
#' mod <- onam(model_formula, list_of_deep_models,
#'             data_train, n_ensemble = 1, epochs = 10)
#' decompose(mod)
#' }
#' @export decompose
decompose <- function(object, data = NULL) {
  if (!require_keras()) {
    invisible(return(NULL))
  }
  if (is.null(data)) {
    effects <- object$feature_effects
  } else {
    if (inherits(object, "onam")) {
      effects <- predict(object, data)$feature_effects
    } else {
      stop(
        "When calling \`decompose\` with a non-default \`data\`
      argument, \`object\` has to be the \`onam\`-model.",
        call. = FALSE
      )
    }
  }
  theta_deep <-
    object$model_info$theta[setdiff(names(object$model_info$theta), "linear")]
  max_order <- max(as.numeric(names(theta_deep)))
  orders <- sort(unique(c(1, 2, as.numeric(
    names(theta_deep)
  ))))
  effect_order_matrix <- matrix(0, nrow = nrow(effects), ncol = length(orders))
  sens_info = rep(0, length(orders))
  names(sens_info) <- orders
  for (idx_order in seq_along(orders)) {
    tmp_name <-
      get_effect_names_for_order(theta_deep[[as.character(orders[idx_order])]],
                                 max_order)
    sens_info[idx_order] <- length(tmp_name)
    tmp_effects <- effects[, tmp_name]
    if (!is.null(dim(tmp_effects))) {
      tmp_effects <- rowSums(tmp_effects)
    }
    effect_order_matrix[, idx_order] <-
      tmp_effects
  }
  # effect_order_matrix <-
  #   matrix(effect_order_matrix[, ncol(effect_order_matrix):1])
  colnames(effect_order_matrix) <- orders
  tmp_var <- stats::var(effect_order_matrix)
  sens_info <- sens_info[length(sens_info):1]
  out <- list(var_decomp = diag(tmp_var) / sum(diag(tmp_var)))
  attr(out, "sens_info") <- sens_info
  attr(out, "target") <- object$model_info$target
  class(out) <- "var_decomp"
  out
}
#' @method print var_decomp
#' @export
print.var_decomp <- function(x, ...) {
  out_txt <- "\nFraction of total variance explained per order"
  if (attr(x, "target") == "binary") {
    out_txt <- paste0(out_txt, " (on logit level)")
  }
  out_txt <- paste0(out_txt, ":")
  cat(out_txt)
  for (i in seq_along(x$var_decomp)) {
    if (i == 1) {
      cat("\nMain effects:", x$var_decomp[i])
    } else {
      cat("\nInteractions of order ",
          names(x$var_decomp)[i],
          ": ",
          round(x$var_decomp[i], 4))
    }
  }
  # cat("\nGeneralized sensitivity indices for each effect:")
  # for (i in seq_along(x$sens_index)) {
  #   cat("\n", names(x$sens_index)[i],
  #       ": ", round(x$sens_index[i], 5))
  # }
  invisible(x)
}
#' Plot variance decomposition
#' @param x Object of class `var_decomp` as returned by [decompose]
#' @param ... further arguments, currently unused
#' @returns Returns a 'ggplot2' object showing the fraction of total
#' variance explained by each interaction order
#' @method plot var_decomp
#' @export
plot.var_decomp <- function(x, ...) {
  orders <- names(x$var_decomp)
  order_labels <- ifelse(orders == "1",
                         "Main effects",
                         paste0("Order ", orders, " interactions"))
  data_plot <- data.frame(
    order = factor(order_labels, levels = order_labels),
    variance = as.numeric(x$var_decomp)
  )
  ylab_val <- "Fraction of explained variance"
  if (attr(x, "target") == "binary") {
    ylab_val <- paste0(ylab_val, " (logit scale)")
  }
  ggplot2::ggplot(data_plot,
                  ggplot2::aes(x = .data$order, y = .data$variance)) +
    ggplot2::geom_bar(stat = "identity") +
    ggplot2::geom_text(
      ggplot2::aes(label = scales::percent(.data$variance, accuracy = 0.1)),
      vjust = -0.3
    ) +
    ggplot2::scale_y_continuous(labels = scales::percent,
                                expand = ggplot2::expansion(mult = c(0, 0.1))) +
    ggplot2::xlab(NULL) +
    ggplot2::ylab(ylab_val) +
    ggplot2::theme(panel.grid.major.x = ggplot2::element_blank())
}
#' Compute Generalized Sobol Indices for dependent features
#' @param object Either model of class `onam` as returned from [onam] or
#' evaluation outcome as returned from [predict.onam].
#' @param data Data for which the model is to be evaluated. If \code{NULL}
#' (DEFAULT), the data from model fitting is used.
#' @returns Returns a named vector of percentage of variance explained by each
#' interaction order.
#' @examplesIf reticulate::py_module_available(tensorflow)
#' \donttest{
#' # Basic example for a simple ONAM-model
#' # Create training data
#' n <- 1000
#' x1 <- runif(n, -2, 2)
#' x2 <- runif(n, -2, 2)
#' y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
#'   pweibull(-x2, shape = 0.5)) +
#'   x1 * x2
#' data_train <- cbind(x1, x2, y)
#' # Define model
#' model_formula <- y ~ mod1(x1) + mod1(x2) +
#'   mod1(x1, x2)
#' mod1 <- function(inputs) {
#'   outputs <- inputs %>%
#'     layer_dense(units = 16, activation = "relu") %>%
#'     layer_dense(units = 8, activation = "linear",
#'                 use_bias = TRUE) %>%
#'     layer_dense(units = 1, activation = "linear",
#'                 use_bias = TRUE)
#'   keras_model(inputs, outputs)
#' }
#' list_of_deep_models <- list(mod1 = mod1)
#' # Fit model
#' mod <- onam(model_formula, list_of_deep_models,
#'             data_train, n_ensemble = 1, epochs = 10)
#' gen_sobol(mod)
#' }
#' @details For details on generalized sobol indices, see
#' Chastaing et al. (2012) \doi{10.1214/12-EJS749}.
#' @export gen_sobol
gen_sobol <- function(object, data = NULL) {
  if (!require_keras()) {
    invisible(return(NULL))
  }
  if (inherits(object, "onam") |
      inherits(object, "onam_prediction"))
  {
    tmp_var <- stats::var(object$feature_effects)
  } else
  {
    stop(
      "`gen_sobol`` can only be called on objects of class
       `onam` and `onam_prediction`.",
      call. = FALSE
    )
  }
  total_var <- sum(tmp_var)
  gen_sobol_index_1 <- diag(tmp_var)
  gen_sobol_index_2 <- (rowSums(tmp_var) - gen_sobol_index_1) / total_var
  gen_sobol_index_1 <- gen_sobol_index_1 / total_var
  total_sobol_index <- gen_sobol_index_1 + gen_sobol_index_2
  effect_orders <- sapply(seq_along(object$model_info$theta), function(i)
    rep(
      names(object$model_info$theta)[i],
      length(object$model_info$theta[[i]])
    )) %>% unlist() %>% as.numeric()
  out_df <-
    data.frame(
      effect_order = effect_orders,
      gen_sobol_index_1 = gen_sobol_index_1,
      gen_sobol_index_2 = gen_sobol_index_2,
      total_sobol_index = total_sobol_index
    )
  out <- list(
    out_df = out_df,
    gen_sobol_index_1 = gen_sobol_index_1,
    gen_sobol_index_2 = gen_sobol_index_2,
    total_sobol_index = total_sobol_index
  )
  attr(out, "target") <- object$model_info$target
  class(out) <- "gen_sobol"
  out
}
#' @method print gen_sobol
#' @export
print.gen_sobol <- function(x, ...) {
  out_txt <- "\nGeneralized Sobol Indices"
  if (attr(x, "target") == "binary") {
    out_txt <- paste0(out_txt, " (on logit level)")
  }
  out_txt <- paste0(out_txt, ":")
  cat(out_txt)
  tmp <- x$out_df %>% dplyr::arrange(.data$effect_order)
  for (i in unique(tmp$effect_order)) {
    sub_tmp <- tmp %>% dplyr::filter(.data$effect_order == i) %>% as.matrix()
    if (i == 1) {
      cat("\nMain effects:")
    } else {
      cat("\nInteractions of order ", i, ": ")
    }
    for (j in seq(nrow(sub_tmp)))
    {
      cat("\n ", rownames(sub_tmp)[j], sub_tmp[j, "total_sobol_index"])
    }
  }
  invisible(x)
}
#' Plot generalized Sobol indices
#' @param x Object of class `gen_sobol` as returned by [gen_sobol]
#' @param ... further arguments, currently unused
#' @returns Returns a 'ggplot2' object showing, for each fitted effect, the
#' generalized Sobol index split into the effect's own contribution
#' (`gen_sobol_index_1`) and its contribution through interactions with
#' other effects of the same order (`gen_sobol_index_2`)
#' @details For details on generalized sobol indices, see
#' Chastaing et al. (2012) \doi{10.1214/12-EJS749}.
#' @method plot gen_sobol
#' @export
plot.gen_sobol <- function(x, ...) {
  data_plot <- x$out_df
  data_plot$effect <- rownames(data_plot)
  data_plot <-
    data_plot[order(data_plot$effect_order, -data_plot$total_sobol_index), ]
  data_plot$effect <- factor(data_plot$effect, levels = data_plot$effect)
  data_plot_long <- data.frame(
    effect = rep(data_plot$effect, 2),
    component = rep(c("Effect contribution",
                      "Effect induced by other terms of the same order"),
                    each = nrow(data_plot)),
    value = c(data_plot$gen_sobol_index_1, data_plot$gen_sobol_index_2)
  )
  ylab_val <- "Generalized Sobol Index"
  if (attr(x, "target") == "binary") {
    ylab_val <- paste0(ylab_val, " (logit scale)")
  }
  ggplot2::ggplot(data_plot_long,
                  ggplot2::aes(x = .data$effect, y = .data$value,
                               fill = .data$component)) +
    ggplot2::geom_bar(stat = "identity") +
    ggplot2::scale_y_continuous(labels = scales::percent) +
    ggplot2::xlab(NULL) +
    ggplot2::ylab(ylab_val) +
    ggplot2::labs(fill = NULL) +
    ggplot2::theme(
      axis.text.x = ggplot2::element_text(angle = 45, hjust = 1),
      legend.position = "top"
    )
}

utils::globalVariables(c("x", "y", "prediction", "x1", "x2", "y1", "y2"))
#' Plot Main Effect
#' @param object Either model of class `onam` as returned from [onam] or
#' model evaluation outcome as returned from [predict.onam]
#' @param feature Feature for which the effect is to be plotted, must be present
#' in the model formula. For interaction terms, use plotInteractionEffect
#' @param reference_level Reference level to be used when plotting categorical
#' effects for better interpretability. Effect of the reference level will be
#' set to zero and subtracted from all over factor levels.
#' @param labs An optional named vector that can contain axis labels.
#' Expected vector names are 'xlab', 'ylab' and 'effect'. If both 'ylab' and
#' 'effect' are specified, 'effect' will be used as y-axis label.
#' @returns Returns a ggplot2 object of the specified effect
#' @examplesIf reticulate::py_module_available(tensorflow)
#' # Basic example for a simple ONAM-model
#' # Create training data
#' n <- 1000
#' x1 <- runif(n, -2, 2)
#' x2 <- runif(n, -2, 2)
#' y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
#'   pweibull(-x2, shape = 0.5)) +
#'   x1 * x2
#' data_train <- cbind(x1, x2, y)
#' # Define model
#' model_formula <- y ~ mod1(x1) + mod1(x2) +
#'   mod1(x1, x2)
#' mod1 <- function(inputs) {
#'   outputs <- inputs %>%
#'     layer_dense(units = 16, activation = "relu") %>%
#'     layer_dense(units = 8, activation = "linear",
#'                 use_bias = TRUE) %>%
#'     layer_dense(units = 1, activation = "linear",
#'                 use_bias = TRUE)
#'   keras_model(inputs, outputs)
#' }
#' list_of_deep_models <- list(mod1 = mod1)
#' # Fit model
#' mod <- onam(model_formula, list_of_deep_models,
#'             data_train, n_ensemble = 1, epochs = 10)
#' plot_main_effect(mod, "x1")
#' @export plot_main_effect
plot_main_effect <-
  function(object, feature, reference_level = NULL, labs = NULL) {
    check_inputs_plot(object,
                      feature,
                      reference_level = reference_level)
    data_plot <-
      data.frame(x = object$data[, feature], y = object$feature_effects[, feature])
    if (feature %in% object$model_info$categorical_features) {
      plt <- plot_main_categorical(data_plot, reference_level)
    } else {
      plt <- ggplot2::ggplot(data_plot, ggplot2::aes(x = x, y = y)) +
        ggplot2::geom_point()
    }
    # Extract optional labels
    xlab_val <- if (!is.null(labs) && "xlab" %in% names(labs)) {
      labs[["xlab"]]
    } else {
      feature
    }
    ylab_val <- if (!is.null(labs) && "effect" %in% names(labs)) {
      labs[["effect"]]
    } else if (!is.null(labs) && "ylab" %in% names(labs)) {
      labs[["ylab"]]
    } else {
      eff_label_helper(object$model_info$target)
    }

    plt + ggplot2::ylab(ylab_val) +
      ggplot2::xlab(xlab_val)
  }
#' Plot Interaction Effect
#' @param object Either model of class `onam` as returned from [onam] or
#' model evaluation outcome as returned from [predict.onam]
#' @param feature1,feature2 Effects to be plotted.
#' @param interpolate If TRUE, values will be interpolated for a smooth plot.
#' If FALSE (default), only observations in the data will be plotted.
#' @param labs An optional named vector that can contain axis labels and the
#' name of the effect. Expected vector names are 'xlab', 'ylab' and 'effect'.
#' @param custom_colors color palette object for the interaction plot. Default
#' is "spectral", returning a color palette based on the spectral theme.
#' @param n_interpolate number of values per coordinate axis to interpolate.
#' Ignored if 'interpolate = FALSE'.
#' @param include_main If TRUE, main effects for features feature1 and feature2
#' will be added to the interaction term to give combined effect of main and
#' interaction effects. Default is FALSE.
#' @returns Returns a 'ggplot2' object of the specified effect interaction
#' @examplesIf reticulate::py_module_available(tensorflow)
#' # Basic example for a simple ONAM-model
#' # Create training data
#' n <- 1000
#' x1 <- runif(n, -2, 2)
#' x2 <- runif(n, -2, 2)
#' y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
#'   pweibull(-x2, shape = 0.5)) +
#'   x1 * x2
#' data_train <- cbind(x1, x2, y)
#' # Define model
#' model_formula <- y ~ mod1(x1) + mod1(x2) +
#'   mod1(x1, x2)
#' mod1 <- function(inputs) {
#'   outputs <- inputs %>%
#'     layer_dense(units = 16, activation = "relu") %>%
#'     layer_dense(units = 8, activation = "linear",
#'                 use_bias = TRUE) %>%
#'     layer_dense(units = 1, activation = "linear",
#'                 use_bias = TRUE)
#'   keras_model(inputs, outputs)
#' }
#' list_of_deep_models <- list(mod1 = mod1)
#' # Fit model
#' mod <- onam(model_formula, list_of_deep_models,
#'             data_train, n_ensemble = 1, epochs = 10)
#' plot_inter_effect(mod, "x1", "x2")
#' @export plot_inter_effect
plot_inter_effect <- function(object,
                              feature1,
                              feature2,
                              interpolate = FALSE,
                              labs = NULL,
                              custom_colors = "spectral",
                              n_interpolate = 200,
                              include_main = FALSE) {
  n_cat_effs <- sum(c(feature1, feature2) %in%
                      object$model_info$categorical_features)
  if ((n_cat_effs > 0) & interpolate) {
    warning("Interaction contains categorical feature. No Interpolation will be performed.")
  }
  inter <- paste(feature1, feature2, sep = "_")
  if (!is.null(check_inputs_plot(object, inter, interaction = 1))) {
    inter <- paste(feature2, feature1, sep = "_")
    if (!is.null(check_inputs_plot(object, inter, interaction = 2))) {
      stop(paste(
        "No interaction effect fitted for ",
        feature1,
        " and ",
        feature2,
        ".",
        sep = ""
      ))
    }
    tmp <- feature1
    feature1 <- feature2
    feature2 <- tmp
  }

  # Check for lab values, falling back to defaults
  xlab_val <- if (!is.null(labs) && "xlab" %in% names(labs)) labs[["xlab"]] else feature1
  ylab_val <- if (!is.null(labs) && "ylab" %in% names(labs)) labs[["ylab"]] else feature2
  effect_val <- if (!is.null(labs) && "effect" %in% names(labs)) {
    labs[["effect"]]
  } else {
    eff_label_helper(object$model_info$target)
  }

  legend_pos <- if (nchar(effect_val) > 10) "top" else "right"
  legend_title <- if (nchar(effect_val) > 10) ggplot2::element_text(hjust = 0.5)
    else ggplot2::element_text()
  colorbar_guide <- if (nchar(effect_val) > 10) {
    ggplot2::guide_colorbar(title.position = "top")
  } else {
    ggplot2::guide_colorbar(title.position = "top", title.vjust = 1)
  }

  tmp_xlab <- ggplot2::xlab(xlab_val)
  tmp_ylab <- ggplot2::ylab(ylab_val)
  if (typeof(custom_colors) != "closure") {
    if (custom_colors == "spectral") {
      custom_colors <-
        grDevices::colorRampPalette(colors = (x = RColorBrewer::brewer.pal(n = 11, name = "Spectral")))
    }
  }
  eff <- object$feature_effects[, inter]
  if (include_main)
  {
    f1_in <- feature1 %in% colnames(object$feature_effects)
    f2_in <- feature2 %in% colnames(object$feature_effects)
    if (!all(f1_in, f2_in))
    {
      warning(
        paste0(
          "`include_main` has been set to `true`, ",
          "but not all features were included as main effects."
        ),
        call. = FALSE
      )
    }
    if (f1_in)
    {
      eff <- eff + object$feature_effects[, feature1]
    }
    if (f2_in)
    {
      eff <- eff + object$feature_effects[, feature2]
    }
  }
  if (interpolate & (n_cat_effs == 0)) {
    tmp_interp <-
      akima::interp(
        x = object$data[, feature1],
        y = object$data[, feature2],
        z = eff,
        nx = n_interpolate,
        ny = n_interpolate,
        duplicate = "mean"
      )
    data_plot <-
      data.frame(
        x = rep(tmp_interp$x, length(tmp_interp$y)),
        y = rep(tmp_interp$y, each = length(tmp_interp$x)),
        prediction = tmp_interp$z %>% c()
      )
    data_plot <- data_plot[which(!is.na(data_plot$prediction)), ]
    aes_gradient <-
      ggplot2::scale_fill_gradientn(
        colors =
          custom_colors(n = 100),
        values =
          scales::rescale(c(
            min(data_plot$prediction),
            mean(data_plot$prediction),
            max(data_plot$prediction)
          )),
        guide = colorbar_guide,
        limits = c(min(data_plot$prediction), max(data_plot$prediction)),
        name = effect_val
      )
    geom_param <- ggplot2::geom_tile()
    aes_param <- ggplot2::aes(x = x, y = y, fill = prediction)
  } else {
    data_plot <-
      data.frame(x = object$data[, feature1],
                 y = object$data[, feature2],
                 prediction = eff)
    aes_gradient <-
      ggplot2::scale_color_gradientn(
        colors =
          custom_colors(n = 100),
        values =
          scales::rescale(c(
            min(data_plot$prediction),
            mean(data_plot$prediction),
            max(data_plot$prediction)
          )),
        guide = colorbar_guide,
        limits = c(min(data_plot$prediction), max(data_plot$prediction)),
        name = effect_val
      )
    geom_param <- ggplot2::geom_point()
    prediction <- NULL #remove cmd check note
    if (n_cat_effs == 1) {
      data_plot <- data_plot %>%
        dplyr::group_by(x, y) %>%
        dplyr::summarise(prediction = mean(prediction), .groups = "drop")
      tmp_ylab <-
        ggplot2::ylab(effect_val)
      if (feature1 %in% object$model_info$categorical_features) {
        aes_gradient <- ggplot2::scale_color_discrete(name = feature1)
        data_plot$x <- as.factor(data_plot$x)
        aes_param <- ggplot2::aes(x = y,
                                  y = prediction,
                                  color = x)
      } else {
        aes_gradient <- ggplot2::scale_color_discrete(name = feature2)
        data_plot$y <- as.factor(data_plot$y)
        tmp_xlab <- ggplot2::xlab(feature1)
        aes_param <- ggplot2::aes(x = x,
                                  y = prediction,
                                  color = y)
      }
    } else {
      aes_param <- ggplot2::aes(x = x, y = y, color = prediction)
    }
  }
  inter_theme <- ggplot2::theme(
    plot.title.position = "plot",
    plot.caption.position =  "plot",
    # plot.margin = grid::unit(c(0,0,0,0), "mm"),
    panel.grid = ggplot2::element_blank(),
    panel.background = ggplot2::element_blank(),
    legend.position = legend_pos,
    legend.title = legend_title,
    legend.box.spacing = ggplot2::unit(0.5, "pt")
  )
  ggplot2::ggplot(data_plot, aes_param) +
    # geom_point(size = 0.75) +
    geom_param +
    aes_gradient +
    inter_theme +
    tmp_xlab + tmp_ylab
}
eff_label_helper <- function(target) {
  if (target == "continuous") {
    "Effect"
  } else if (target == "binary") {
    "Effect on logit scale"
  }
}
plot_main_categorical <-
  function(data_plot, reference_level = NULL) {
    x_uq <- unique(data_plot$x)
    first_x <-
      sapply(x_uq, function(x_val)
        min(which(data_plot$x == x_val)))
    y_uq <- data_plot$y[first_x]
    data_plot_categ <-
      data.frame(x = x_uq, y = y_uq)
    if (!is.null(reference_level))
    {
      x_ref_idx <- which(data_plot_categ$x == reference_level)
      data_plot_categ$y <- data_plot_categ$y -
        data_plot_categ$y[x_ref_idx]
      data_plot_categ$x[x_ref_idx] <-
        paste0(reference_level, " (Reference)")
    }
    ggplot2::ggplot(data_plot_categ, ggplot2::aes(x = x, y = y)) +
      ggplot2::geom_bar(stat = "identity") +
      ggplot2::geom_text(ggplot2::aes(
        y = 0,
        label = x,
        vjust = 0.5 + 0.75 * sign(y)
      )) +
      ggplot2::theme(
        axis.text.x = ggplot2::element_blank(),
        axis.ticks.x = ggplot2::element_blank()
      )
  }

test_that("Input checks work", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  x4 <- runif(n, -2, 2)
  y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3), pweibull(-x2, shape = 0.5)) +
    dt(x3, 1) * 4 +
    x1 * x2
  data_train <- cbind(x1, x2, x3, x4, y)
  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3)
  f2 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3) + mod1(x5)
  f3 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3) + mod4(x2)
  f4 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  expect_error(
    onam(
      f2,
      list_of_deep_models,
      data_train,
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = "Feature\\(s\\) x5 in formula, but not present in data. Make sure the features align with colnames\\(data\\).",
  )
  expect_error(
    onam(
      f3,
      list_of_deep_models,
      data_train,
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = "Model formula contains model mod4, but mod4 is not supplied in 'list_of_deep_models'.",
  )
  expect_error(
    onam(
      f1,
      list_of_deep_models,
      data_train,
      categorical_features = c("x5"),
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = "x5 provided in categorical_features, but not present in data. Make sure the features align with colnames\\(data\\).",
  )
  expect_warning(
    onam(
      f4,
      list_of_deep_models,
      data_train,
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = "Features in lower order effects do not appear in higher order effects. We recommend fitting a residual term that includes all lower order terms.",
  )
  expect_warning(
    onam(
      f1,
      list_of_deep_models,
      data_train,
      categorical_features = c("x4"),
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = "Feature\\(s\\) x4 stated as categorical, but not present in model formula.",
  )
  mod <- onam(
    f1,
    list_of_deep_models,
    data_train,
    n_ensemble = 1,
    epochs = 10,
    progresstext = FALSE,
    verbose = 0
  )
  expect_error(plot_main_effect(mod, "x4"),
               "x4 is not present in the fitted model effects.")
  expect_error(plot_inter_effect(mod, "x1", "x3"),
               "x3_x1 is not present in the fitted model effects.")
  expect_warning(
    plot_main_effect(mod, "x1", "1"),
    "Reference level given, but x1 is not modeled as categorical."
  )
})

#Get submodel until penultimate layer
get_intermediate_model <-
  function(model,
           idx_layer = NULL) {
    if (is.null(idx_layer)) {
      idx_layer <- length(model$layers) - 1
    }
    deep_part_in <-
      keras3::get_layer(model, index = as.integer(idx_layer))
    keras3::keras_model(model$input,
                        deep_part_in$output)
  }
#Solve linear system after removing linear dependencies
solve_singular_matrix <- function(tmp_u, pivot) {
  tmp_u_reduced <- tmp_u[, pivot]
  tryCatch(
    tmp <- solve(t(tmp_u_reduced) %*% tmp_u_reduced),
    error = function(error)
    {
      pivot <<-
        solve_singular_matrix(tmp_u,
                              pivot[1:(length(pivot) - 1)])
    }
  )
  return(pivot)
}
#get list of all submodel indices in model_info$theta
get_model_idx_list <- function(model_info) {
  lapply(seq_along(model_info$theta),
         function(idx_order_inter)
           lapply(seq_along(model_info$theta[[idx_order_inter]]),
                  function(x)
                    c(idx_order_inter, x))) %>%
    unlist(recursive = FALSE)
}
#get penultimate output for all submodels
get_u <- function(model_list,
                  model_idx_list,
                  model_info,
                  data) {
  data_dictionary <- get_data_dictionary(model_info)
  get_bias_helper <- function(model) {
    utils::tail(model$layers, 1)[[1]]$use_bias
  }
  u_list <-
    lapply(seq_along(model_idx_list),
           function(idx) {
             input <- model_list[[model_idx_list[[idx]]]] %>%
               get_intermediate_model() %>%
               stats::predict(data[[data_dictionary[[model_idx_list[[idx]]]]]],
                              verbose = 0)
             if (get_bias_helper(model_list[[model_idx_list[[1]]]])) {
               input <- cbind(input, 1)
             }
             return(input)
           })
  u_dims <-
    lapply(seq_along(u_list),
           function(idx_model)
             ncol(u_list[[idx_model]]))
  u <- unlist(u_list) %>%
    matrix(ncol =
             sum(u_dims %>% unlist()))
  u <- cbind(u, 1)
  u_idx_list <- list(1:u_dims[[1]])
  for (idx in seq_along(u_dims)[-1]) {
    idx_last <- u_idx_list[[idx - 1]][u_dims[[idx - 1]]]
    u_idx_list[[idx]] <- (idx_last + 1):(idx_last + u_dims[[idx]])
  }
  list(u = u,
       u_dims = u_dims,
       u_idx_list = u_idx_list)

}
#get list of weights in last layer of each subfunction
get_w_list <- function(model_list,
                       model_idx_list,
                       model_info,
                       u_idx_list) {
  w_list_sep <- lapply(seq_along(model_idx_list),
                       function(idx) {
                         weights <- model_list[[model_idx_list[[idx]]]] %>%
                           keras3::get_layer(index = -1) %>%
                           keras3::get_weights()
                         return(unlist(weights))
                       })
  w <- c(unlist(w_list_sep), 0)
  lapply(seq_along(w_list_sep),
         function(idx_model) {
           tmp_w <- w
           tmp_w[-u_idx_list[[idx_model]]] <- 0
           return(tmp_w)
         })
}
#get model orders based on model index list
get_model_order <- function(model_idx_list) {
  lapply(model_idx_list,
         function(idx_model)
           idx_model[1]) %>% unlist()
}
#post hoc orthogonalization of fitted submodels
pho <- function(model_list, model_info, data_fit) {
  model_idx_list <- get_model_idx_list(model_info)
  u_object <-
    get_u(model_list, model_idx_list, model_info, data_fit)
  w_list <-
    get_w_list(model_list, model_idx_list, model_info,
               u_object$u_idx_list)
  w_list_old <- w_list
  model_order <- get_model_order(model_idx_list)
  #Iterate over interaction depth
  for (idx_ortho in 2:(length(model_info$theta))) {
    if (length(model_info$theta) == 1) {
      break
    }
    list_idx_order_lower <- which(model_order >= idx_ortho)
    idx_rel <- u_object$u_idx_list[list_idx_order_lower] %>%
      unlist()
    tmp_u <- u_object$u
    tmp_u[,-idx_rel] <- 0
    tmp_u[, ncol(tmp_u)] <- 1
    h <- crossprod(tmp_u)
    qr_res <- qr(h)
    h_order <- qr_res$pivot[1:qr_res$rank]
    pivot <- solve_singular_matrix(tmp_u, h_order)
    tmp_u_reduced <- tmp_u[, pivot]
    tmp_inverse <- solve(t(tmp_u_reduced) %*% tmp_u_reduced)
    list_idx_order_higher <- which(model_order == (idx_ortho - 1))
    outputs_list <-
      lapply(list_idx_order_higher,
             function(idx_order_higher) {
               return(u_object$u %*%
                        w_list[[idx_order_higher]])
             })
    z_list <-
      lapply(outputs_list,
             function(outputs) {
               tmp_z <- tmp_inverse %*% t(tmp_u_reduced) %*%
                 outputs
               z <- rep(0, ncol(tmp_u))
               z[pivot] <- tmp_z
               z
             })
    for (idx_order_higher in seq_along(list_idx_order_higher)) {
      idx_model <- list_idx_order_higher[[idx_order_higher]]
      w_list[[idx_model]] <-
        w_list[[idx_model]] -
        z_list[[idx_order_higher]]
    }
    for (idx_order_lower in list_idx_order_lower) {
      tmp_weight_update <-
        rep(0, u_object$u_dims[[idx_order_lower]])
      for (idx_order_higher in seq_along(list_idx_order_higher)) {
        tmp_weight_update <-
          tmp_weight_update +
          z_list[[idx_order_higher]][u_object$u_idx_list[[idx_order_lower]]]
      }
      weight_update <- rep(0, ncol(tmp_u))
      weight_update[u_object$u_idx_list[[idx_order_lower]]] <-
        tmp_weight_update
      w_list[[idx_order_lower]] <-
        w_list[[idx_order_lower]] + weight_update
    }
  }
  allOutputMeans <-
    lapply(seq_along(model_idx_list),
           function(idx_model) {
             u_object$u %*%
               w_list[[idx_model]] %>% mean()
           })
  w_list <-
    lapply(seq_along(w_list),
           function(w_Idx) {
             tmp_w <- w_list[[w_Idx]]
             tmp_w[length(tmp_w)] <- if (w_Idx == 1)
               sum(unlist(allOutputMeans)[-1])
             else
               - allOutputMeans[[w_Idx]]
             tmp_w
           })
  list(
    model_list = model_list,
    w_list = w_list,
    u_dims = u_object$u_dims,
    w_list_old = w_list_old
  )
}
show_progress <- function(i, n_ensemble) {
  cat('\r', paste0("Fitting model ", i, " of ", n_ensemble))
  utils::flush.console()
}
#post hoc orthogonalization of fitted (and  separately orthogonalized) ensemble
#members
pho_ensemble <- function(data_model_eval, model_info) {
  n_ensemble <- max(data_model_eval$data_predictions$model)
  data <- data_model_eval$data
  n <- nrow(data)
  model_idx_list <- get_model_idx_list(model_info)
  model_order <- get_model_order(model_idx_list)
  n_models <- length(model_order)
  u <- unlist(data_model_eval$predictions_features_ensemble) %>%
    matrix(nrow = n)
  w <- diag(1, nrow = n_models)
  #Iterate over interaction depth
  for (idx_ortho in 2:(length(model_info$theta))) {
    list_idx_order_lower <- which(model_order >= idx_ortho)
    tmp_u <- u
    tmp_u[,-list_idx_order_lower] <- 0
    h <- crossprod(tmp_u)
    qr_res <- qr(h)
    h_order <- qr_res$pivot[1:qr_res$rank]
    pivot <- solve_singular_matrix(tmp_u, h_order)
    tmp_u_reduced <- tmp_u[, pivot]
    tmp_inverse <- solve(t(tmp_u_reduced) %*% tmp_u_reduced)
    list_idx_order_higher <- which(model_order == (idx_ortho - 1))
    outputs_list <-
      lapply(list_idx_order_higher,
             function(idx_order_higher) {
               return(u %*% w[, idx_order_higher])
             })
    z_list <-
      lapply(outputs_list,
             function(outputs) {
               tmp_z <- tmp_inverse %*% t(tmp_u_reduced) %*% outputs
               z <- rep(0, n_models)
               z[pivot] <- tmp_z
               return(z)
             })
    for (idx_order_higher in seq_along(list_idx_order_higher)) {
      idx_model <- list_idx_order_higher[[idx_order_higher]]
      w[, idx_model] <-
        w[, idx_model] -
        z_list[[idx_order_higher]]
    }
    for (idx_order_lower in list_idx_order_lower) {
      tmp_weight_update <- 0
      for (idx_order_higher in seq_along(list_idx_order_higher)) {
        tmp_weight_update <-
          tmp_weight_update +
          z_list[[idx_order_higher]][idx_order_lower]
      }
      w[idx_order_lower, idx_order_lower] <-
        w[idx_order_lower, idx_order_lower] + tmp_weight_update
    }
  }
  outputs_post_ensemble <-
    u %*% w
  colnames(outputs_post_ensemble) <-
    names(data_model_eval$predictions_features_ensemble)
  list(w, outputs_post_ensemble)
}
#' @importFrom dplyr %>%
#' @export
NULL

#Define DNN architecture for simulation
get_submodel <- function(inputs, regularizer = NULL) {
  outputs <- inputs %>%
    # layer_dense(
    #   units = 256,
    #   activation = "relu",
    #   use_bias = TRUE,
    #   kernel_regularizer = regularizer
    # ) %>%
    keras3::layer_dense(
      units = 128,
      activation = "relu",
      use_bias = TRUE,
      kernel_regularizer = regularizer
    ) %>%
    keras3::layer_dense(
      units = 64,
      activation = "relu",
      use_bias = TRUE,
      kernel_regularizer = regularizer,
      dtype = "float64"
    ) %>%
    keras3::layer_dense(
      units = 32,
      activation = "relu",
      use_bias = TRUE,
      kernel_regularizer = regularizer,
      dtype = "float64"
    ) %>%
    keras3::layer_dense(
      units = 16,
      activation = "relu",
      use_bias = TRUE,
      kernel_regularizer = regularizer,
      dtype = "float64"
    ) %>%
    keras3::layer_dense(
      units = 8,
      activation = "linear",
      use_bias = TRUE,
      kernel_regularizer = regularizer,
      dtype = "float64"
    ) %>%
    keras3::layer_dense(units = 1,
                        activation = "linear",
                        use_bias = TRUE)
  keras3::keras_model(inputs, outputs)
}
#Define linear Model DNN structure
get_linear_submodel <- function(inputs) {
  outputs <- inputs %>%
    keras3::layer_dense(
      units = 1,
      activation = "linear",
      use_bias = FALSE,
      dtype = "float64"
    )
  keras3::keras_model(inputs, outputs)
}
#Define model architecture for categorical variable
get_categorical_submodel <- function(inputs) {
  outputs <- inputs %>%
    keras3::layer_dense(
      units = 1,
      activation = "linear",
      use_bias = TRUE,
      dtype = "float64"
    )
  keras3::keras_model(inputs, outputs)
}
#Derive Theta from model Formula
get_theta <-
  function(model_formula,
           list_of_deep_models,
           feature_names,
           categorical_features,
           target) {
    #Separate Symbols
    all_feature_indic <- FALSE
    theta_list <- lapply(model_formula, find_symbol)
    outcome_var <- theta_list[[2]][[1]]
    parts_list <- list()
    while (length(theta_list) > 0) {
      tmp_item <- theta_list[[length(theta_list)]]
      idx_to_remove <- c(length(theta_list))
      while (length(tmp_item) > 1 &
             any(unlist(lapply(tmp_item, is.list)))) {
        idx_to_remove <- c(idx_to_remove, length(tmp_item))
        tmp_item <- tmp_item[[length(tmp_item)]]
        if (length(tmp_item) > 1) {
          if (!is.list(tmp_item[[2]])) {
            if (as.character(tmp_item[[2]]) == ".") {
              all_feature_indic <- TRUE
              tmp_item[[2]] <- NULL
              outcome_idx <- which(feature_names == outcome_var)
              tmp_item <-
                if (length(outcome_idx) > 0) {
                  c(tmp_item, feature_names[-outcome_idx])
                } else {
                  c(tmp_item, feature_names)
                }
            }
          }
        }
      }
      parts_list <- c(parts_list, list(tmp_item))
      theta_list[[idx_to_remove]] <- NULL
    }
    idx_sign <- which(lapply(parts_list, function(item) {
      all(lapply(item, function(sub_item) {
        sub_item == as.name("+") |
          sub_item == as.name("~") |
          sub_item == as.name(outcome_var)
      }) %>%
        unlist())
    }) %>% unlist())
    parts_list[idx_sign] <- NULL
    # idx_additive_comps <- which(lapply(parts_list, function(item) {
    #   any(lapply(item, function(sub_item) {
    #     sub_item == as.name("+")
    #   }) %>%
    #     unlist()) | length(item) == 1
    # }) %>% unlist())
    # additive_comps_long <- parts_list[idx_additive_comps]
    # parts_list[idx_additive_comps] <- NULL
    check_inputs_formula(
      parts_list,
      list_of_deep_models,
      feature_names,
      categorical_features,
      outcome_var
    )
    # additive_comps <- lapply(additive_comps_long, function(item) {
    #   if (length(item) == 1) {
    #     item
    #   } else {
    #     item[which(lapply(item,
    #                       function(sub_item)
    #                         sub_item != as.name("+")) %>%
    #                  unlist())]
    #   }
    # }) %>% unlist() %>% unique()
    order_inter <-
      lapply(parts_list, function(item)
        length(item) - 1) %>%
      unlist()
    order_highest <- max(order_inter)
    ordered_parts_list <-
      parts_list[(1:length(parts_list))[order(order_inter, decreasing = TRUE)]]
    order_counts <- table(order_inter)
    order_counts <-
      order_counts[order(as.numeric(names(order_counts)), decreasing = TRUE)]
    idx_counts <- list()
    start_index <- 1
    for (i in seq_along(order_counts)) {
      end_index <- start_index + order_counts[i] - 1
      seq_list <- start_index:end_index
      idx_counts[[i]] <- seq_list
      start_index <- end_index + 1
    }
    orders_unique <- as.numeric(names(order_counts))
    theta_deep_name <-
      lapply(seq_along(orders_unique), function(countIdx)
        ordered_parts_list[idx_counts[[countIdx]]])
    idx_names <-
      lapply(seq_along(theta_deep_name), function(idx_item) {
        lapply(seq_along(theta_deep_name[[idx_item]]), function(idx_item_in) {
          lapply(seq_along(theta_deep_name[[idx_item]][[idx_item_in]]), function(idx_item_in_2) {
            if (as.character(theta_deep_name[[idx_item]][[idx_item_in]][[idx_item_in_2]])
                %in% names(list_of_deep_models)) {
              return(c(idx_item, idx_item_in, idx_item_in_2))
            }
          })
        })
      }) %>%
      unlist()
    idx_names <-
      idx_names %>% split(rep(1:(length(idx_names) / 3), each = 3))
    theta_deep <- theta_deep_name
    model_list <- list()
    for (item in idx_names) {
      name = as.character(theta_deep[[item]])
      theta_deep[[item]] <- NULL
      if (length(model_list) < item[1]) {
        model_list[[item[1]]] <- list()
      }
      if (length(model_list[[item[1]]]) < item[2]) {
        model_list[[item[1]]][[item[2]]] <- name
      } else {
        model_list[[item[1]]][[item[2]]] <- c(model_list[[item[2]]], name)
      }
    }
    names(theta_deep) <- orders_unique
    names(model_list) <- orders_unique
    # theta <- c(theta_deep, linear = list(additive_comps))
    theta <- theta_deep
    list(
      theta = theta,
      name_models = model_list,
      categorical_features = categorical_features,
      outcome = outcome_var,
      target = target,
      all_feature_indic = all_feature_indic,
      model_formula = model_formula
    )
  }
#help function to detect symbols
find_symbol <- function(list_current) {
  lapply(list_current, function(item) {
    if (is.symbol(item)) {
      item
    } else {
      find_symbol(item)
    }
  })
}
#create ONAM model based on model info and deep model list
create_model <- function(model_info,
                         list_of_deep_models,
                         categorical_features,
                         cat_counts,
                         target = "continuous",
                         learning_rate) {
  inputs_list <- create_inputs(model_info, categorical_features, cat_counts)
  model_list <-
    create_models(model_info, inputs_list, list_of_deep_models)
  model_whole <-
    compile_model(inputs_list,
                  model_list,
                  target = target,
                  learning_rate = learning_rate)
  list(model = model_whole, model_list = model_list)
}
#create inputs for each submodel
create_inputs <- function(model_info,
                          categorical_features,
                          cat_counts) {
  # if (length(model_info$theta$linear) > 0) {
  #   inputs_linear <- lapply(1:length(model_info$theta$linear),
  #                           function(x)
  #                             keras3::layer_input(shape = 1))
  # } else {
  #   inputs_linear <- NULL
  # }
  # inputs_deep <-
  #   lapply(1:length(model_info$theta[setdiff(names(model_info$theta),
  #                                            "linear")]),
  #          function(idx_inter_count) {
  #            lapply(model_info$theta[[idx_inter_count]],
  #                   function(theta_sub) {
  #                     n_inputs <- length(theta_sub)
  #                     for (feature_name in theta_sub) {
  #                       if (as.character(feature_name) %in% names(cat_counts)) {
  #                         n_inputs <- n_inputs + cat_counts[[feature_name]] - 1
  #                       }
  #                     }
  #                     keras3::layer_input(shape = n_inputs)
  #                   })
  #          })
  inputs <-
    lapply(1:length(model_info$theta), function(idx_inter_count) {
      lapply(model_info$theta[[idx_inter_count]], function(theta_sub) {
        n_inputs <- length(theta_sub)
        for (feature_name in theta_sub) {
          if (as.character(feature_name) %in% names(cat_counts)) {
            n_inputs <- n_inputs + cat_counts[[feature_name]] - 1
          }
        }
        keras3::layer_input(shape = n_inputs)
      })
    })
  names(inputs) <-
    names(model_info$theta)
  inputs
}
#create all submodels
create_models <-
  function(model_info,
           inputs_list,
           list_of_deep_models) {
    models <- list()
    for (idx_p in
         names(model_info$theta)) {
      models[[idx_p]] <-
        lapply(seq_along(model_info$name_models[[idx_p]]), function(idx_model) {
          name_model <- model_info$name_models[[idx_p]][[idx_model]]
          return(list_of_deep_models[[name_model]](inputs_list[[idx_p]][[idx_model]]))
        })
    }
    models
  }
#concatenate models in model_list
concatenate_model_list <-
  function(model_list,
           bias = FALSE,
           activation = "linear") {
    concatenated <-
      keras3::layer_concatenate(lapply(model_list, function(model)
        model$output))
    dense_layer <-
      keras3::layer_dense(
        units = 1,
        use_bias = bias,
        trainable = FALSE,
        activation = activation
      )
    output <- dense_layer(concatenated)
    tmp_weights <- dense_layer$get_weights()
    tmp_weights[[1]] <- matrix(rep(1, length(model_list)), ncol = 1)
    if (bias) {
      tmp_weights[[2]] <- tmp_weights[[2]] - tmp_weights[[2]]
    }
    dense_layer$set_weights(tmp_weights)
    output
  }
#compile created pho ensemble member
compile_model <-
  function(inputs_list,
           model_list,
           target = "continuous",
           learning_rate) {
    submodels <-
      unlist(model_list, use.names = FALSE) #Why does this matter?
    all_inputs <- unlist(inputs_list, use.names = FALSE)
    target_activation <-
      if (target == "continuous") {
        "linear"
      } else {
        "sigmoid"
      }
    model_whole <- submodels %>%
      concatenate_model_list(bias = TRUE, activation = target_activation)
    model_whole <- keras3::keras_model(all_inputs, model_whole)
    loss <- if (target == "continuous") {
      keras3::loss_mean_squared_error()
    } else if (target == "binary") {
      keras3::loss_binary_crossentropy()
    } else {
      stop()
    }
    model_whole %>%
      keras3::compile(
        loss = loss,
        optimizer = keras3::optimizer_adam(learning_rate = learning_rate)
      )

  }
#prepare data for model fitting by bringing it into the right input dimensions
prepare_data <- function(data_original,
                         model_info,
                         categorical_features) {
  data_additive <-
    lapply(model_info$theta$linear, function(feature_name) {
      data_original[, as.character(feature_name)]
    })
  deep_data <-
    lapply(unlist(model_info$theta[setdiff(names(model_info$theta), "linear")], recursive = FALSE), function(theta_sub) {
      ret_mat <- matrix(nrow = nrow(data_original), ncol = 0)
      for (feature in theta_sub) {
        feature <- as.character(unlist(feature))
        ret_mat <-
          cbind(ret_mat, if (feature %in% categorical_features) {
            encode_dummy(data_original[, feature], feature)
          } else {
            data_original[, feature]
          })
        if (!feature %in% categorical_features) {
          colnames(ret_mat)[ncol(ret_mat)] <- feature
        }
      }
      return(ret_mat)
    })
  newData <- c(deep_data, data_additive)
  unname(newData)
}
#Generate a display/column name for a single effect.
get_effect_name <- function(features, max_order) {
  order <- length(features)
  if (order == max_order) {
    paste0("Interaction ", order)
  } else {
    paste(features, collapse = "_")
  }
}
#Generate display names for every effect in one order group (an element
#of model_info$theta).
get_effect_names_for_order <- function(theta_order, max_order) {
  vapply(theta_order, function(effect) {
    get_effect_name(unlist(effect), max_order)
  }, character(1))
}
#Generate display names for every fitted effect, in the same order
#model_info$theta is iterated in evaluate_onam_single().
get_all_effect_names <- function(model_info) {
  max_order <- max(as.numeric(names(model_info$theta)))
  unlist(lapply(model_info$theta, get_effect_names_for_order,
                max_order = max_order))
}
#Create "dictionary" yielding data indices for nested model list
get_data_dictionary <- function(model_info) {
  dictionary_list <- list()
  idx_tmp <- 0
  for (i1 in seq_along(model_info$theta)) {
    dictionary_list[[i1]] <- list()
    for (i2 in seq_along(model_info$theta[[i1]])) {
      idx_tmp <- idx_tmp + 1
      dictionary_list[[i1]][[i2]] <- idx_tmp
    }
  }
  dictionary_list
}
#create dummy variables for categorical data
encode_dummy <- function(x, name) {
  uq_vals <- unique(x)
  ret_mat <- matrix(sapply(uq_vals[-1], function(val) {
    x == val
  }), nrow = length(x))
  colnames(ret_mat) <- paste(name, uq_vals[-1], sep = "_")
  ret_mat
}
#get number of unique categories per categorical feature - 1 (because of dummy
#encoding)
get_category_counts <- function(categorical_features, data) {
  ret_list <-
    lapply(categorical_features, function(feature) {
      length(unique(data[, feature])) - 1
    })
  names(ret_list) <- categorical_features
  ret_list
}
get_output <-
  function(model,
           prediction_function,
           model_data,
           data,
           target,
           model_info) {
    if (!is.null(model)) {
      if (is.null(prediction_function)) {
        for (tmp_class in class(model)) {
          prediction_function <-
            utils::getS3method("predict", tmp_class, optional = TRUE)
          if (!is.null(prediction_function))
            break
        }
        if (is.null(prediction_function)) {
          stop(
            paste0(
              "Model of class ",
              class(model),
              " supplied without `prediction_function`, but no S3 method for class ",
              class(model),
              "exists. Please specify a prediction function that returns a vector of predictions."
            ),
            call. = FALSE
          )
        }
      }
      if (is.null(model_data)) {
        model_data <- data
      }
      y <- prediction_function(model, model_data)
      if (inherits(y, "factor")) {
        if (is.null(target)) {
          warning("Prediction function supplied factor. Binary outcome is assumed.")
          target <- "binary"
          y <- as.numeric(as.character(y))
        } else if (target != "binary") {
          stop(
            paste0(
              "Prediction function supplied factor, but target was supplied as ",
              target,
              "."
            ),
            call. = FALSE
          )
        } else {
          y <- as.numeric(as.character(y))
        }
      }
      if (!is.vector(y)) {
        if ("predictions" %in% names(y)) {
          y <- y[["predictions"]]
        } else {
          stop(
            "Prediction function does not return an appropriate outcome. Please specify a prediction function that returns a vector of predictions.",
            call. = FALSE
          )
        }
      }

    } else {
      outcome_index <- which(colnames(data) ==
                               as.character(model_info$outcome))
      if (length(outcome_index) != 1) {
        stop(
          paste0(
            "Outcome ",
            model_info$outcome,
            " specified in model formula, but not present in data."
          ),
          call. = FALSE
        )
      } else {
        y <- data[, outcome_index]
      }
    }
    y
  }

#' Build a stacked dense (DNN) model constructor
#'
#' Returns a *function* `function(inputs) {...}` — not a model itself —
#' that builds a feed-forward stack of `layer_dense()` calls on top of
#' whatever `inputs` tensor it is later called with, and returns the
#' resulting `keras3::keras_model()`.
#'
#' Layers can be specified either as a simple vector of unit counts, or
#' as a data frame / tibble giving full per-layer control over
#' activation, use_bias, initializers, regularizers, layer name, etc.
#'
#' @param units Optional integer vector of layer sizes. Ignored if
#'   `layer_df` is supplied. Every layer gets `activation = "relu"` and
#'   `use_bias = TRUE`.
#' @param layer_df Optional data frame / tibble with one row per layer.
#'   Must contain a `units` column. May also contain any of:
#'   \itemize{
#'     \item `activation` (character)
#'     \item `use_bias` (logical)
#'     \item `kernel_initializer`, `bias_initializer` (character, or a
#'       list-column of initializer objects, e.g.
#'       `keras3::initializer_he_normal()`)
#'     \item `kernel_regularizer`, `bias_regularizer`,
#'       `activity_regularizer` (list-column; each element is either
#'       `NULL` or an object like `keras3::regularizer_l2(0.01)`)
#'     \item `name` (character, layer name)
#'   }
#'   Any column not supplied falls back to `layer_dense()`'s own default.
#'   Columns holding actual keras objects (regularizers, and
#'   initializers if you don't just want a string) must be *list-columns*
#'   so each row can carry a different object or NULL, e.g.:
#'   \preformatted{
#'   tibble::tibble(
#'     units = c(64, 32, 1),
#'     activation = c("relu", "relu", "linear"),
#'     kernel_regularizer = list(keras3::regularizer_l2(0.001), NULL, NULL)
#'   )
#'   }
#' @param default_activation Activation function to be used if not otherwise
#' specified in `layer_df`.
#'
#' @return A function of a single argument `inputs` (a keras input
#'   tensor) that returns a `keras3::keras_model()`.
#'
#' @examplesIf reticulate::py_module_available(tensorflow)
#' mod_units <- build_dnn(units = c(16, 8, 1))
#' mod_df <- build_dnn(layer_df = data.frame(
#'   units = c(32, 16, 8, 1),
#'   activation = c("relu", "relu", "relu", "linear"),
#'   use_bias = c(FALSE, FALSE, FALSE, TRUE)
#' ))
#' @export
build_dnn <- function(units = NULL,
                      layer_df = NULL,
                      default_activation = NULL) {
  if (is.null(units) && is.null(layer_df)) {
    stop("Either `units` or `layer_df` must be supplied.")
  }
  if (!is.null(units) && !is.null(layer_df)) {
    stop("Supply only one of `units` or `layer_df`, not both.")
  }
  if (is.null(default_activation))
  {
    default_activation <- "relu"
  }
  # --- Mode 1: plain units vector --------------------------------------
  if (!is.null(units)) {
    layer_df <- data.frame(
      units      = as.integer(units),
      activation = default_activation,
      use_bias   = TRUE,
      stringsAsFactors = FALSE
    )
  }

  if (!"units" %in% names(layer_df)) {
    stop("`layer_df` must contain a `units` column.")
  }
  if (layer_df$units[nrow(layer_df)] != 1)
  {
    warning(
      "Last layer must have exactly 1 output unit. An additional layer with 1 output unit was added."
    )
    layer_df <- rbind(layer_df, layer_df[-1, ])
    layer_df$units[nrow(layer_df)] <- 1
  }
  if (!is.null(units)) {
    layer_df$activation[nrow(layer_df)] <- "linear"
  }
  # --- fill in missing optional columns with sensible defaults --------
  scalar_defaults <- list(activation = default_activation, use_bias   = TRUE)
  # columns that may hold real keras objects -> always kept as list-columns
  object_cols <- c(
    "kernel_initializer",
    "bias_initializer",
    "kernel_regularizer",
    "bias_regularizer",
    "activity_regularizer"
  )

  n <- nrow(layer_df)

  for (col in names(scalar_defaults)) {
    if (!col %in% names(layer_df))
      layer_df[[col]] <- scalar_defaults[[col]]
  }
  for (col in object_cols) {
    if (!col %in% names(layer_df))
      layer_df[[col]] <- vector("list", n)
  }
  if (!"name" %in% names(layer_df))
    layer_df[["name"]] <- NA_character_
  if (layer_df$activation[nrow(layer_df)] != "linear")
  {
    warning("Last layer must have linear activation. Activation in last layer was set to linear.")
    layer_df$activation[nrow(layer_df)] <- "linear"
  }
  # helper: get row-i value whether column is a list-column or atomic
  get_val <- function(df, col, i) {
    if (!col %in% names(df))
      return(NULL)
    v <- df[[col]]
    if (is.list(v))
      v[[i]]
    else
      v[i]
  }

  # --- return the model constructor -------------------------------------
  function(inputs) {
    outputs <- inputs
    for (i in seq_len(n)) {
      args <- list(
        object     = outputs,
        units      = layer_df$units[[i]],
        activation = get_val(layer_df, "activation", i),
        use_bias   = get_val(layer_df, "use_bias", i)
      )

      for (col in object_cols) {
        val <- get_val(layer_df, col, i)
        if (!is.null(val) &&
            !(length(val) == 1 && all(is.na(val)))) {
          args[[col]] <- val
        }
      }

      nm <- get_val(layer_df, "name", i)
      if (!is.null(nm) && !is.na(nm))
        args$name <- nm

      outputs <- do.call(keras3::layer_dense, args)
    }

    keras3::keras_model(inputs, outputs)
  }
}

test_that("Orthogonalization works", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  x4 <- runif(n, -2, 2)
  y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3), pweibull(-x2, shape = 0.5)) +
    dt(x3, 1) * 4 +
    x1 * x2
  data_train <- cbind(x1, x2, x3, x4, y)
  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  mod <- onam(
    f1,
    list_of_deep_models,
    data_train,
    n_ensemble = 1,
    epochs = 10,
    progresstext = FALSE,
    verbose = 0
  )
  eff_mat <- mod$feature_effects
  #Mean Centering works
  expect_all_equal(colMeans(eff_mat[, 2:ncol(eff_mat)]), 0)
  #Effects are orthogonal
  var_mat <- var(eff_mat)
  expect_all_equal(c(var_mat[-1, 1], var_mat[3:nrow(var_mat), 2]), 0)
})
test_that("Seed works", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  y <- 2 * x1 - 2 * x2
  data_train <- cbind(x1, x2, y)

  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(.)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  expect_error(
    mod <- onam(
      f1,
      list_of_deep_models,
      data_train,
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      seed = 1,
      verbose = 0
    ),
    regexp = NA
  )
  tmp_eff <- mod$feature_effects[, "x1"]
  mod2 <- onam(
    f1,
    list_of_deep_models,
    data_train,
    n_ensemble = 1,
    epochs = 10,
    progresstext = FALSE,
    seed = 1,
    verbose = 0
  )
  expect_equal(mod2$feature_effects[, "x1"], tmp_eff)
})
extract_dense_specs <- function(model) {
  dense_layers <- Filter(
    function(l) grepl("Dense", class(l)[1], ignore.case = TRUE),
    model$layers
  )
  lapply(dense_layers, function(l) {
    cfg <- l$get_config()
    list(
      units      = cfg$units,
      activation = cfg$activation,
      use_bias   = cfg$use_bias
    )
  })
}
test_that("build_dnn works", {
  skip_if_no_conda_env()
  inputs <- layer_input(shape = c(10))
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16,
                  activation = "relu",
                  use_bias = TRUE) %>%
      layer_dense(units = 8,
                  activation = "relu",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  built_mod1 <- mod1(inputs)
  mod2 <- build_dnn(units = c(16, 8, 1))
  built_mod2 <- mod2(inputs)
  reference_specs <- extract_dense_specs(built_mod1)
  built_specs     <- extract_dense_specs(built_mod2)
  #Same number of layers
  expect_length(built_specs, length(reference_specs))
  # Same units / activation / use_bias, layer by layer
  expect_equal(built_specs, reference_specs)
  # Shapes and parameter counts
  expect_equal(built_mod2$output_shape, built_mod1$output_shape)
  expect_equal(built_mod2$count_params(), built_mod1$count_params())
  # DF parametrization
  mod3 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 32,
                  activation = "relu",
                  use_bias = FALSE) %>%
      layer_dense(units = 16,
                  activation = "relu",
                  use_bias = FALSE) %>%
      layer_dense(units = 8,
                  activation = "relu",
                  use_bias = FALSE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  built_mod3 <- mod3(inputs)
  mod4 <- build_dnn(layer_df = data.frame(
    units = c(32, 16, 8, 1),
    activation = c("relu", "relu", "relu", "linear"),
    use_bias = c(FALSE, FALSE, FALSE, TRUE)
  ))
  built_mod4 <- mod4(inputs)
  reference_specs <- extract_dense_specs(built_mod3)
  built_specs     <- extract_dense_specs(built_mod4)
  #Same number of layers
  expect_length(built_specs, length(reference_specs))
  # Same units / activation / use_bias, layer by layer
  expect_equal(built_specs, reference_specs)
  # Shapes and parameter counts
  expect_equal(built_mod2$output_shape, built_mod1$output_shape)
  expect_equal(built_mod2$count_params(), built_mod1$count_params())
})

#' @importFrom rlang .data
#' @importFrom stats predict
evaluate_onam_single <- function(model,
                                 data,
                                 model_info,
                                 categorical_features) {
  data_fit <- prepare_data(data, model_info,
                           categorical_features)
  model_idx_list <- get_model_idx_list(model_info)
  u_object <-
    get_u(model$model_list, model_idx_list, model_info, data_fit)
  predictions_submodel <-
    lapply(model$w_list,
           function(w)
             u_object$u %*% w)
  predictions_submodel_old <-
    lapply(model$w_list_old,
           function(w)
             u_object$u %*% w)
  model_names <- get_all_effect_names(model_info)
  names(predictions_submodel) <- model_names
  names(predictions_submodel_old) <- model_names
  list(predictions_submodel = predictions_submodel,
       predictions_submodel_old = predictions_submodel_old)
}
#' Evaluate orthogonal neural additive model
#' @docType methods
#' @param object model of class `onam` as returned from [onam] to be
#' evaluated
#' @param newdata Data for which the model is to be evaluated. If NULL (default),
#' data with which \code{model} was fitted is used.
#' @returns Returns a list containing data, model output for each observation in
#' `newdata` and main and interaction effects obtained by the model
#' @param ... some methods for this generic require additional arguments. None
#' are used in this method.
#' @aliases predict
#' @method predict onam
#' @export
predict.onam <- function(object, newdata = NULL, ...) {
  if (!require_keras()) {
    invisible(return(NULL))
  }
  if (is.null(newdata)) {
    newdata <- object$data
  }
  model_info <- object$model_info
  n <- nrow(newdata)
  n_ensemble <- length(object$ensemble)
  categorical_features <- object$categorical_features
  predictions_separate <-
    lapply(
      object$ensemble,
      evaluate_onam_single,
      data = newdata,
      model_info = model_info,
      categorical_features = categorical_features
    )
  effect_names <- names(predictions_separate[[1]][[1]])
  n_effects <- length(effect_names)
  data_predictions <-
    data.frame(
      y = unlist(predictions_separate),
      effect =
        rep(rep(effect_names,
                each = n),
            2 * n_ensemble),
      onam = rep(rep(c("after", "before"),
                     each = n * n_effects),
                 n_ensemble),
      observation = rep(1:n, n_effects * 2 * n_ensemble),
      model = rep(1:n_ensemble, each = n * 2 * n_effects)
    )
  predictions_features_ensemble <-
    lapply(effect_names,
           function(effect_name) {
             data_predictions %>%
               dplyr::filter(.data$onam == "after",
                             .data$effect == effect_name) %>%
               dplyr::group_by(.data$observation) %>%
               dplyr::summarise(totaleffect = mean(.data$y)) %>%
               dplyr::select(.data$totaleffect) %>% unlist()
           })
  names(predictions_features_ensemble) <-
    effect_names
  predictions_features <-
    unlist(predictions_features_ensemble) %>%
    matrix(nrow = n) %*%
    object$w_post_ensemble
  colnames(predictions_features) <-
    effect_names
  predictions_total <-
    data_predictions %>%
    dplyr::filter(.data$onam == "after") %>%
    dplyr::group_by(.data$observation) %>%
    dplyr::summarise(prediction = sum(.data$y) / n_ensemble) %>%
    dplyr::select(.data$prediction) %>% unlist()
  out <- list(
    data = newdata,
    predictions = predictions_total,
    feature_effects = predictions_features,
    model_info = model_info
  )
  class(out) <- "onam_prediction"
  out
}
evaluate_onam_pre <- function(model_list) {
  data <- model_list$data
  model_info <- model_list$model_info
  categorical_features <- model_list$categorical_features
  n <- nrow(data)
  n_ensemble <- length(model_list$ensemble)
  predictions_separate <-
    lapply(
      model_list$ensemble,
      evaluate_onam_single,
      data = data,
      model_info = model_info,
      categorical_features = categorical_features
    )
  effect_names <- names(predictions_separate[[1]][[1]])
  n_effects <- length(effect_names)
  data_predictions <-
    data.frame(
      y = unlist(predictions_separate),
      effect =
        rep(rep(effect_names,
                each = n),
            2 * n_ensemble),
      onam = rep(rep(c("after", "before"),
                     each = n * n_effects),
                 n_ensemble),
      observation = rep(1:n, n_effects * 2 * n_ensemble),
      model = rep(1:n_ensemble, each = n * 2 * n_effects)
    )
  predictions_features_ensemble <-
    lapply(effect_names,
           function(effect_name) {
             data_predictions %>%
               dplyr::filter(.data$onam == "after",
                             .data$effect == effect_name) %>%
               dplyr::group_by(.data$observation) %>%
               dplyr::summarise(totaleffect = mean(.data$y)) %>%
               dplyr::select(.data$totaleffect) %>% unlist()
           })
  names(predictions_features_ensemble) <-
    effect_names
  predictions_total <-
    data_predictions %>%
    dplyr::filter(.data$onam == "after") %>%
    dplyr::group_by(.data$observation) %>%
    dplyr::summarise(prediction = sum(.data$y) / n_ensemble) %>%
    dplyr::select(.data$prediction) %>% unlist()
  list(
    data = data,
    predictions_total = predictions_total,
    data_predictions = data_predictions,
    predictions_features_ensemble = predictions_features_ensemble
  )
}

#' Fit orthogonal neural additive model
#'
#' @description
#' Fits an interpretable neural additive model with post hoc orthogonalization
#' for a given network architecture and user-specified feature sets.
#'
#' @param formula Formula for model fitting. Specify deep parts with the same
#' name as `list_of_deep_models`.
#' @param list_of_deep_models List of named models used in `model_formula`.
#' @param data Data to be fitted
#' @param model Prediction model that is to be explained. Output of the model as
#' returned from `prediction_function(model)` will be used as model output. If
#' `NULL`(default), the outcome has to be present in `data`.
#' @param prediction_function Prediction function to be used to generate the
#' outcome. Only used if `model` is specified. If `NULL`(default), S3-method
#' based on the `model` argument is used.
#' @param model_data Data used for generating predictions of `model`. Necessary
#' for some models that require specific data formats, i.e. xgboost.
#' If `NULL`(default), `data` is used. Only used if `model` is specified.
#' @param categorical_features Vector of feature names of categorical features.
#' @param target Target of prediction task. Can be either "continuous" or
#' "binary". For "continuous"(default), an additive model for the prediction of
#' a continuous outcome is fitted. For "binary", a binary classification with
#' sigmoid activation in the last layer is fitted.
#' @param epochs Number of epochs to train the model. See
#' \code{\link[keras3]{fit}} for details.
#' @param learning_rate Learning rate for model fitting. See
#' \code{\link[keras3]{compile}} for details.
#' @param n_ensemble Number of orthogonal neural additive model ensembles
#' @param callback Callback to be called during training. See
#' \code{\link[keras3]{fit}} for details.
#' @param seed Random seed used by R, python, numpy, and backend framework.
#' See \code{\link[keras3]{set_random_seed}} for details.
#' @param progresstext Show model fitting progress. If `TRUE`, shows current
#' number of ensemble being fitted
#' @param verbose Verbose argument for internal model fitting. Used for
#' debugging. See \code{\link[keras3]{fit}} for
#' details.
#' @importFrom keras3 fit
#' @returns Returns a model object of class \code{onam}, containing all ensemble
#' members, ensemble weights, and main and interaction effect outputs.
#' @examplesIf reticulate::py_module_available(tensorflow)
#' # Basic example for a simple ONAM-model
#' # Create training data
#' n <- 1000
#' x1 <- runif(n, -2, 2)
#' x2 <- runif(n, -2, 2)
#' y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
#'   pweibull(-x2, shape = 0.5)) +
#'   x1 * x2
#' data_train <- cbind(x1, x2, y)
#' # Define model
#' model_formula <- y ~ mod1(x1) + mod1(x2) +
#'   mod1(x1, x2)
#' mod1 <- function(inputs) {
#'   outputs <- inputs %>%
#'     layer_dense(units = 16, activation = "relu") %>%
#'  layer_dense(units = 8, activation = "linear",
#'              use_bias = TRUE) %>%
#'  layer_dense(units = 1, activation = "linear",
#'              use_bias = TRUE)
#'  keras_model(inputs, outputs)
#' }
#' list_of_deep_models <- list(mod1 = mod1)
#' # Fit model
#' mod <- onam(model_formula, list_of_deep_models,
#'             data_train, n_ensemble = 1, epochs = 10)
#' summary(mod)
#' @details For more details see
#' Koehler et al. (2025) <doi:10.1038/s44387-025-00033-7>.
#' @export onam
onam <- function(formula,
                 list_of_deep_models,
                 data,
                 model = NULL,
                 prediction_function = NULL,
                 model_data = NULL,
                 categorical_features = NULL,
                 target = "continuous",
                 n_ensemble = 10,
                 epochs = 500,
                 learning_rate = 0.001,
                 callback = NULL,
                 seed = NULL,
                 progresstext = FALSE,
                 verbose = 0) {
  if (!require_keras()) {
    invisible(return(NULL))
  }
  inputs <- as.list(environment())
  check_inputs_onam(inputs)
  if (!is.null(seed))
  {
    keras3::set_random_seed(seed)
  }
  feature_names <- colnames(data)
  model_info <-
    get_theta(formula,
              list_of_deep_models,
              feature_names,
              categorical_features,
              target)
  cat_counts <- get_category_counts(categorical_features, data)
  y <-
    get_output(model,
               prediction_function,
               model_data,
               data,
               target,
               model_info)
  check_y_features(data, y, model_info, target)
  data_fit <-
    prepare_data(data, model_info, categorical_features)
  ensemble <- list()
  for (i in 1:n_ensemble) {
    if (progresstext) {
      show_progress(i, n_ensemble)
    }
    model_object <-
      create_model(
        model_info,
        list_of_deep_models,
        categorical_features,
        cat_counts,
        target,
        learning_rate
      )
    model_whole <- model_object$model
    model_list <- model_object$model_list
    #Fit model####
    # callback <-
    #   keras3::keras$callbacks$EarlyStopping(monitor = "loss",
    #                                        patience = 10)
    history <- model_whole %>%
      keras3::fit(
        data_fit,
        y,
        epochs = epochs,
        callbacks = callback,
        verbose = verbose
      )
    #Orthogonalize####
    ensemble[[i]] <-
      pho(model_list, model_info, data_fit)
  }
  model_list_pho <- list(
    ensemble = ensemble,
    model_info = model_info,
    data = data,
    categorical_features = categorical_features
  )
  data_model_eval <-
    evaluate_onam_pre(model_list_pho)
  pho_ensemble_list <- pho_ensemble(data_model_eval, model_info)
  w_post_ensemble <- pho_ensemble_list[[1]]
  outputs_post_ensemble <- pho_ensemble_list[[2]]
  out <- c(
    model_list_pho,
    call = match.call(),
    w_post_ensemble = list(w_post_ensemble),
    feature_effects = list(outputs_post_ensemble),
    predictions = list(rowSums(outputs_post_ensemble)),
    y = list(y)
  )
  class(out) <- "onam"
  out
}
#' Get summary of an onam object
#'
#' @description
#' generates a summary of a fitted `onam` object including
#' information on ensembling strategy and performance metrics such as
#' correlation and degree of interpretabiltity
#'
#' @param object onam object of class `onam` as returned from
#' \code{\link{onam}} to be summarized
#' @param x object of class \code{\link{summary.onam}}.
#' @param ... further arguments passed to or from other methods.
#' @returns Gives summary of the `onam` object, including model inputs, number
#' of ensembles, correlation of model output and original outcome variable, and
#' interpretability metrics i_1 and i_2
#' @details
#' For examples see \code{\link[=onam]{example(onam)}}
#' @method summary onam
#' @export
summary.onam <- function(object, ...) {
  var_decomp <- decompose(object)$var_decomp

  if (object$model_info$target == "continuous") {
    convergence_metric <- stats::cor(object$predictions, object$y)
    conv_met_attr <- "cor"
  } else {
    if (is.numeric(object$y) && !is.integer(object$y))
    {
      convergence_metric <- stats::cor(object$predictions, object$y)
      conv_met_attr <- "cor_p"
    } else
    {
      convergence_metric <- pROC::auc(object$y, object$predictions)
      conv_met_attr <- "auc"
    }
    # stats::cor(object$predictions,
    #            object$y, method = "spearman")
  }
  res <- list(
    call = object$call,
    formula = object$model_info$model_formula,
    n_ensemble = length(object$ensemble),
    conv_metric = convergence_metric,
    i_1 = var_decomp["1"],
    i_2 = var_decomp["2"],
    degree_expl = sum(var_decomp[c("1", "2")])
  )
  class(res) <- "summary.onam"
  attr(res, "target") <- object$model_info$target
  attr(res, "conv_met") <- conv_met_attr
  res
}

#' @rdname summary.onam
#' @method print summary.onam
#' @export
print.summary.onam <- function(x, ...) {
  cat("\nCall:\n")
  print(x$call)
  cat("\nFormula:\n")
  print(x$formula)
  if (attr(x, "conv_met") == "cor") {
    cat(
      "\nCorrelation of model prediction with outcome variable: ",
      round(x$conv_metric, 4),
      sep = ""
    )
  } else if (attr(x, "conv_met") == "cor_p") {
    cat(
      "\nCorrelation of onam probabilities with original model predicted probabilities: ",
      round(x$conv_metric, 4),
      sep = ""
    )
  } else  if (attr(x, "conv_met") == "auc") {
    cat("\nPrediction AUC: ", round(x$conv_metric, 4), sep = "")
  }
  cat("\nNumber of ensemble members: ", x$n_ensemble)
  cat("\nI_1: ",
      round(x$i_1, digits = 4),
      "; I_2:  ",
      round(x$i_2, digits = 4),
      sep = "")
  cat("\nDegree of interpretability: ",
      round(x$degree_expl, digits = 4),
      sep = "")
  invisible(x)
}

test_that("ONAM works", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  x4 <- runif(n, -2, 2)
  y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3), pweibull(-x2, shape = 0.5)) +
    dt(x3, 1) * 4 +
    x1 * x2
  data_train <- cbind(x1, x2, x3, x4, y)
  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  expect_error(
    onam(
      f1,
      list_of_deep_models,
      data_train,
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = NA
  )
})
test_that("ONAM predict works", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  x4 <- runif(n, -2, 2)
  y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3), pweibull(-x2, shape = 0.5)) +
    dt(x3, 1) * 4 +
    x1 * x2
  data_train <- cbind(x1, x2, x3, x4, y)
  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(x1, x2, x3)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  mod <- onam(
    f1,
    list_of_deep_models,
    data_train,
    n_ensemble = 1,
    epochs = 10,
    progresstext = FALSE,
    verbose = 0
  )
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  x4 <- runif(n, -2, 2)
  newdat = cbind(x1, x2, x3, x4)
  expect_false(isTRUE(all.equal((predict(mod, newdat))$predictions, mod$predictions
  )))
})
test_that("ONAM works with model", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  y <- 2 * x1 - 2 * x2
  data_train <- cbind(x1, x2, y)

  simple_mod <- lm(y ~ ., as.data.frame(data_train))

  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(.)
  f2 <- outcome ~ mod1(x1) + mod1(.)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  expect_error(
    mod <- onam(
      f1,
      list_of_deep_models,
      data_train,
      simple_mod,
      model_data = as.data.frame(data_train),
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = NA
  )
  expect_warning(
    onam(
      f2,
      list_of_deep_models,
      data_train,
      simple_mod,
      model_data = as.data.frame(data_train),
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = "Model formula includes term of type `deep_model\\(.\\)`. Data contains column y, which has a correlation of 1 with the outcome. If using a term like `deep_model\\(.\\)` and supplying a `model` to generate the response, make sure that the original outcome is not contained in `data`."
  )
})
test_that("Binary Outcome Works", {
  skip_if_no_conda_env()
  n <- 100
  x1 <- runif(n, -2, 2)
  x2 <- runif(n, -2, 2)
  x3 <- runif(n, -2, 2)
  eta = sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3), pweibull(-x2, shape = 0.5)) +
    dt(x3, 1) * 4 +
    x1 * x2
  y <- runif(n) < 1 / (1 + exp(eta))
  data_train <- cbind(x1, x2, x3, y)
  # Define model
  f1 <- y ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(.)
  mod1 <- function(inputs) {
    outputs <- inputs %>%
      layer_dense(units = 16, activation = "relu") %>%
      layer_dense(units = 8,
                  activation = "linear",
                  use_bias = TRUE) %>%
      layer_dense(units = 1,
                  activation = "linear",
                  use_bias = TRUE)
    keras_model(inputs, outputs)
  }
  list_of_deep_models <- list(mod1 = mod1)
  # Fit model
  expect_error(
    mod <- onam(
      f1,
      list_of_deep_models,
      data_train,
      target = "binary",
      n_ensemble = 1,
      epochs = 10,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = NA
  )
  glm_dat <- as.data.frame(data_train)
  logmod <- glm(y ~ x1 + x2 + x3,
                data = glm_dat,
                family = binomial())
  f2 <- glm_out ~ mod1(x1) + mod1(x2) + mod1(x3) +
    mod1(x1, x2) + mod1(.)
  expect_error(
    mod_glm <- onam(
      f2,
      list_of_deep_models,
      glm_dat,
      logmod,
      target = "binary",
      n_ensemble = 1,
      epochs = 100,
      progresstext = FALSE,
      verbose = 0
    ),
    regexp = NA
  )
  expect_warning(
    summary(mod_glm),
    regexp = NA
  )
})

#check inputs for mismatch between formula and features/models
#' @keywords internal
check_inputs_formula <-
  function(parts_list,
           list_of_deep_models,
           feature_names,
           categorical_features,
           outcome_var) {
    features_ls <- lapply(parts_list, function(x) {
      if (!as.character(x[[1]]) %in% names(list_of_deep_models)) {
        stop(
          paste0(
            "Model formula contains model ",
            x[[1]],
            ", but ",
            x[[1]],
            " is not supplied in 'list_of_deep_models'."
          ),
          call. = FALSE
        )
      }
      return(as.character(x[-1]))
    })
    features <- unlist(features_ls)
    all_symbols <- c(features)
    missing_features <-
      all_symbols[which(!all_symbols %in% feature_names)]
    if (length(missing_features)) {
      stop(
        paste0(
          "Feature(s) ",
          paste(missing_features, collapse = ", "),
          " in formula, but not present in data. Make sure the features align with colnames(data)."
        ),
        call. = FALSE
      )
    }
    if (any(!categorical_features %in% feature_names)) {
      missing_features <-
        categorical_features[which(!categorical_features %in% feature_names)]
      stop(
        paste0(
          paste(missing_features, collapse = ", "),
          " provided in categorical_features, but not present in data. Make sure the features align with colnames(data)."
        ),
        call. = FALSE
      )
    }
    cat_not_in_formula <-
      categorical_features[which(!categorical_features %in% all_symbols)]
    if (length(cat_not_in_formula)) {
      warning(
        paste0(
          "Feature(s) ",
          paste(cat_not_in_formula, collapse = ", "),
          " stated as categorical, but not present in model formula."
        ),
        call. = FALSE
      )
    }
    highest_term_idx <- which.max(lapply(features_ls, length))
    highest_term_features <- features_ls[[highest_term_idx]]
    if (any(unlist(lapply(features_ls[-highest_term_idx], function(terms) {
      any(!terms %in% highest_term_features)
    })))) {
      warning(
        "Features in lower order effects do not appear in higher order effects. We recommend fitting a residual term that includes all lower order terms.",
        call. = FALSE
      )
    }
  }
#check inputs for proper object specification and if specified effects were fitted
#' @keywords internal
check_inputs_plot <- function(object,
                              effect,
                              interaction = 0,
                              reference_level = NULL) {
  if (inherits(object, "onam_prediction") |
      inherits(object, "onam")) {
    names <- colnames(object$feature_effects)
  } else {
    stop(
      "Visualization functions can only be called on objects of type 'onam'
         or 'onam_prediction'.",
      call. = FALSE
    )
  }
  if (!effect %in% names) {
    if (interaction == 1) {
      return(interaction)
    } else {
      stop(paste0(effect,
                  " is not present in the fitted model effects."),
           call. = FALSE)
    }
  }
  if (!is.null(reference_level) && (interaction == 0))
  {
    if (!effect %in% object$categorical_features)
    {
      warning(
        paste0(
          "Reference level given, but ",
          effect,
          " is not modeled as categorical."
        ),
        call. = FALSE
      )
    }
    else if (!  reference_level %in%
             unique(object$data[, effect]))
    {
      stop(paste0(reference_level,
                  " is not a level of feature ",
                  effect, "."),
           call. = FALSE)
    }
  }
}
#check inputs of model fitting call
#' @keywords internal
check_inputs_onam <- function(inputs) {
  if (!inherits(inputs$formula, "formula")) {
    stop("Parameter `formula` must be a formula object of the kind `y~x`.",
         call. = FALSE)
  }
  if (!is.list(inputs$list_of_deep_models) |
      is.null(names(inputs$list_of_deep_models)) |
      any(unlist(lapply(inputs$list_of_deep_models, function(x) {
        !inherits(x, "function")
      })))) {
    stop("Parameter `list_of_deep_models` must be a named list of DNN architectures.",
         call. = FALSE)
  }
  if ((!is.data.frame(inputs$data)) && (!is.matrix(inputs$data))) {
    stop("Parameter `data` must be either of class `matrix` or of class `data.frame`",
         call. = FALSE)
  }
  if (!inputs$target %in% c("continuous", "binary")) {
    stop("Parameter `target` must be either `continuous` or `binary`.",
         call. = FALSE)
  }
  if (!(inputs$n_ensemble %% 1 == 0) | inputs$n_ensemble < 0) {
    stop("Parameter `n_ensemble` must be a positive integer.",
         call. = FALSE)
  }
  if (!(inputs$epochs %% 1 == 0) | inputs$epochs < 0) {
    stop("Parameter `epochs` must be a positive integer.",
         call. = FALSE)
  }
  if (is.na(as.logical(inputs$progresstext)))
    stop("Parameter `progresstext` must be `TRUE` or `FALSE`.",
         call. = FALSE)
}
check_y_features <- function(data, y, model_info, target) {
  if (model_info$all_feature_indic & target == "continuous") {
    data_no_outcome <-
      data[, !(colnames(data) %in% as.character(model_info$outcome))]
    f_y_cors <- apply(data_no_outcome, 2, function(x) {
      if (!is.numeric(x)) {
        0
      } else {
        stats::cor(x, y)
      }
    })
    cor_orders <- order(f_y_cors, decreasing = TRUE)
    if (f_y_cors[cor_orders[1]] > 0.99) {
      warning(
        paste0(
          "Model formula includes term of type `deep_model(.)`. Data contains column ",
          colnames(data_no_outcome)[cor_orders[1]],
          ", which has a correlation of ",
          round(f_y_cors[cor_orders[1]], 4),
          " with the outcome. If using a term like `deep_model(.)` and supplying a `model` to generate the response, make sure that the original outcome is not contained in `data`."
        ),
        call. = FALSE
      )
    }
  }
}
#' @importFrom reticulate py_available
require_keras <- function() {
  try(reticulate::use_condaenv("r-keras", required = TRUE),
      silent = TRUE)
  if (!reticulate::py_available(initialize = TRUE))
  {
    message(
      "No Python Environemt available. Use ",
      "keras3::install_keras(method = \"conda\", conda = \"auto\") ",
      "to install recommended environment. If `install_keras` does not work",
      "(esp. on windows machines), use ",
      "install_conda_env()."
    )
    return(FALSE)
  }
  if (!reticulate::py_module_available("tensorflow"))
  {
    message(
      "Tensorflow not available. Use install_keras(method = \"conda\", conda = \"auto\")."
    )
    return(FALSE)
  }
  TRUE
}
#' @keywords internal
skip_if_no_conda_env <- function(envname = "r-keras") {
  testthat::skip_on_cran()

  conda_bin <- tryCatch(
    reticulate::conda_binary(),
    error = function(e)
      NULL
  )

  if (is.null(conda_bin)) {
    testthat::skip("Conda is not available")
  }

  envs <- tryCatch(
    reticulate::conda_list(conda = conda_bin)$name,
    error = function(e)
      character()
  )

  if (!envname %in% envs) {
    testthat::skip(paste0("Conda environment '", envname, "' is not available"))
  }

  reticulate::use_condaenv(envname, required = TRUE)

  reticulate::py_available(initialize = TRUE)

  invisible(TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluation.R
\docType{methods}
\name{predict.onam}
\alias{predict.onam}
\alias{predict}
\title{Evaluate orthogonal neural additive model}
\usage{
\method{predict}{onam}(object, newdata = NULL, ...)
}
\arguments{
\item{object}{model of class \code{onam} as returned from \link{onam} to be
evaluated}

\item{newdata}{Data for which the model is to be evaluated. If NULL (default),
data with which \code{model} was fitted is used.}

\item{...}{some methods for this generic require additional arguments. None
are used in this method.}
}
\value{
Returns a list containing data, model output for each observation in
\code{newdata} and main and interaction effects obtained by the model
}
\description{
Evaluate orthogonal neural additive model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onam.R
\name{onam}
\alias{onam}
\title{Fit orthogonal neural additive model}
\usage{
onam(
  formula,
  list_of_deep_models,
  data,
  model = NULL,
  prediction_function = NULL,
  model_data = NULL,
  categorical_features = NULL,
  target = "continuous",
  n_ensemble = 10,
  epochs = 500,
  learning_rate = 0.001,
  callback = NULL,
  seed = NULL,
  progresstext = FALSE,
  verbose = 0
)
}
\arguments{
\item{formula}{Formula for model fitting. Specify deep parts with the same
name as \code{list_of_deep_models}.}

\item{list_of_deep_models}{List of named models used in \code{model_formula}.}

\item{data}{Data to be fitted}

\item{model}{Prediction model that is to be explained. Output of the model as
returned from \code{prediction_function(model)} will be used as model output. If
\code{NULL}(default), the outcome has to be present in \code{data}.}

\item{prediction_function}{Prediction function to be used to generate the
outcome. Only used if \code{model} is specified. If \code{NULL}(default), S3-method
based on the \code{model} argument is used.}

\item{model_data}{Data used for generating predictions of \code{model}. Necessary
for some models that require specific data formats, i.e. xgboost.
If \code{NULL}(default), \code{data} is used. Only used if \code{model} is specified.}

\item{categorical_features}{Vector of feature names of categorical features.}

\item{target}{Target of prediction task. Can be either "continuous" or
"binary". For "continuous"(default), an additive model for the prediction of
a continuous outcome is fitted. For "binary", a binary classification with
sigmoid activation in the last layer is fitted.}

\item{n_ensemble}{Number of orthogonal neural additive model ensembles}

\item{epochs}{Number of epochs to train the model. See
\code{\link[keras3]{fit}} for details.}

\item{learning_rate}{Learning rate for model fitting. See
\code{\link[keras3]{compile}} for details.}

\item{callback}{Callback to be called during training. See
\code{\link[keras3]{fit}} for details.}

\item{seed}{Random seed used by R, python, numpy, and backend framework.
See \code{\link[keras3]{set_random_seed}} for details.}

\item{progresstext}{Show model fitting progress. If \code{TRUE}, shows current
number of ensemble being fitted}

\item{verbose}{Verbose argument for internal model fitting. Used for
debugging. See \code{\link[keras3]{fit}} for
details.}
}
\value{
Returns a model object of class \code{onam}, containing all ensemble
members, ensemble weights, and main and interaction effect outputs.
}
\description{
Fits an interpretable neural additive model with post hoc orthogonalization
for a given network architecture and user-specified feature sets.
}
\details{
For more details see
Koehler et al. (2025) \url{doi:10.1038/s44387-025-00033-7}.
}
\examples{
\dontshow{if (reticulate::py_module_available(tensorflow)) withAutoprint(\{ # examplesIf}
# Basic example for a simple ONAM-model
# Create training data
n <- 1000
x1 <- runif(n, -2, 2)
x2 <- runif(n, -2, 2)
y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
  pweibull(-x2, shape = 0.5)) +
  x1 * x2
data_train <- cbind(x1, x2, y)
# Define model
model_formula <- y ~ mod1(x1) + mod1(x2) +
  mod1(x1, x2)
mod1 <- function(inputs) {
  outputs <- inputs \%>\%
    layer_dense(units = 16, activation = "relu") \%>\%
 layer_dense(units = 8, activation = "linear",
             use_bias = TRUE) \%>\%
 layer_dense(units = 1, activation = "linear",
             use_bias = TRUE)
 keras_model(inputs, outputs)
}
list_of_deep_models <- list(mod1 = mod1)
# Fit model
mod <- onam(model_formula, list_of_deep_models,
            data_train, n_ensemble = 1, epochs = 10)
summary(mod)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onam.R
\name{summary.onam}
\alias{summary.onam}
\alias{print.summary.onam}
\title{Get summary of an onam object}
\usage{
\method{summary}{onam}(object, ...)

\method{print}{summary.onam}(x, ...)
}
\arguments{
\item{object}{onam object of class \code{onam} as returned from
\code{\link{onam}} to be summarized}

\item{...}{further arguments passed to or from other methods.}

\item{x}{object of class \code{\link{summary.onam}}.}
}
\value{
Gives summary of the \code{onam} object, including model inputs, number
of ensembles, correlation of model output and original outcome variable, and
interpretability metrics i_1 and i_2
}
\description{
generates a summary of a fitted \code{onam} object including
information on ensembling strategy and performance metrics such as
correlation and degree of interpretabiltity
}
\details{
For examples see \code{\link[=onam]{example(onam)}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_main_effect}
\alias{plot_main_effect}
\title{Plot Main Effect}
\usage{
plot_main_effect(object, feature, reference_level = NULL, labs = NULL)
}
\arguments{
\item{object}{Either model of class \code{onam} as returned from \link{onam} or
model evaluation outcome as returned from \link{predict.onam}}

\item{feature}{Feature for which the effect is to be plotted, must be present
in the model formula. For interaction terms, use plotInteractionEffect}

\item{reference_level}{Reference level to be used when plotting categorical
effects for better interpretability. Effect of the reference level will be
set to zero and subtracted from all over factor levels.}

\item{labs}{An optional named vector that can contain axis labels.
Expected vector names are 'xlab', 'ylab' and 'effect'. If both 'ylab' and
'effect' are specified, 'effect' will be used as y-axis label.}
}
\value{
Returns a ggplot2 object of the specified effect
}
\description{
Plot Main Effect
}
\examples{
\dontshow{if (reticulate::py_module_available(tensorflow)) withAutoprint(\{ # examplesIf}
# Basic example for a simple ONAM-model
# Create training data
n <- 1000
x1 <- runif(n, -2, 2)
x2 <- runif(n, -2, 2)
y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
  pweibull(-x2, shape = 0.5)) +
  x1 * x2
data_train <- cbind(x1, x2, y)
# Define model
model_formula <- y ~ mod1(x1) + mod1(x2) +
  mod1(x1, x2)
mod1 <- function(inputs) {
  outputs <- inputs \%>\%
    layer_dense(units = 16, activation = "relu") \%>\%
    layer_dense(units = 8, activation = "linear",
                use_bias = TRUE) \%>\%
    layer_dense(units = 1, activation = "linear",
                use_bias = TRUE)
  keras_model(inputs, outputs)
}
list_of_deep_models <- list(mod1 = mod1)
# Fit model
mod <- onam(model_formula, list_of_deep_models,
            data_train, n_ensemble = 1, epochs = 10)
plot_main_effect(mod, "x1")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var_decomposition.R
\name{plot.gen_sobol}
\alias{plot.gen_sobol}
\title{Plot generalized Sobol indices}
\usage{
\method{plot}{gen_sobol}(x, ...)
}
\arguments{
\item{x}{Object of class \code{gen_sobol} as returned by \link{gen_sobol}}

\item{...}{further arguments, currently unused}
}
\value{
Returns a 'ggplot2' object showing, for each fitted effect, the
generalized Sobol index split into the effect's own contribution
(\code{gen_sobol_index_1}) and its contribution through interactions with
other effects of the same order (\code{gen_sobol_index_2})
}
\description{
Plot generalized Sobol indices
}
\details{
For details on generalized sobol indices, see
Chastaing et al. (2012) \doi{10.1214/12-EJS749}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var_decomposition.R
\name{gen_sobol}
\alias{gen_sobol}
\title{Compute Generalized Sobol Indices for dependent features}
\usage{
gen_sobol(object, data = NULL)
}
\arguments{
\item{object}{Either model of class \code{onam} as returned from \link{onam} or
evaluation outcome as returned from \link{predict.onam}.}

\item{data}{Data for which the model is to be evaluated. If \code{NULL}
(DEFAULT), the data from model fitting is used.}
}
\value{
Returns a named vector of percentage of variance explained by each
interaction order.
}
\description{
Compute Generalized Sobol Indices for dependent features
}
\details{
For details on generalized sobol indices, see
Chastaing et al. (2012) \doi{10.1214/12-EJS749}.
}
\examples{
\dontshow{if (reticulate::py_module_available(tensorflow)) withAutoprint(\{ # examplesIf}
\donttest{
# Basic example for a simple ONAM-model
# Create training data
n <- 1000
x1 <- runif(n, -2, 2)
x2 <- runif(n, -2, 2)
y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
  pweibull(-x2, shape = 0.5)) +
  x1 * x2
data_train <- cbind(x1, x2, y)
# Define model
model_formula <- y ~ mod1(x1) + mod1(x2) +
  mod1(x1, x2)
mod1 <- function(inputs) {
  outputs <- inputs \%>\%
    layer_dense(units = 16, activation = "relu") \%>\%
    layer_dense(units = 8, activation = "linear",
                use_bias = TRUE) \%>\%
    layer_dense(units = 1, activation = "linear",
                use_bias = TRUE)
  keras_model(inputs, outputs)
}
list_of_deep_models <- list(mod1 = mod1)
# Fit model
mod <- onam(model_formula, list_of_deep_models,
            data_train, n_ensemble = 1, epochs = 10)
gen_sobol(mod)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_architecture.R
\name{build_dnn}
\alias{build_dnn}
\title{Build a stacked dense (DNN) model constructor}
\usage{
build_dnn(units = NULL, layer_df = NULL, default_activation = NULL)
}
\arguments{
\item{units}{Optional integer vector of layer sizes. Ignored if
\code{layer_df} is supplied. Every layer gets \code{activation = "relu"} and
\code{use_bias = TRUE}.}

\item{layer_df}{Optional data frame / tibble with one row per layer.
Must contain a \code{units} column. May also contain any of:
\itemize{
\item \code{activation} (character)
\item \code{use_bias} (logical)
\item \code{kernel_initializer}, \code{bias_initializer} (character, or a
list-column of initializer objects, e.g.
\code{keras3::initializer_he_normal()})
\item \code{kernel_regularizer}, \code{bias_regularizer},
\code{activity_regularizer} (list-column; each element is either
\code{NULL} or an object like \code{keras3::regularizer_l2(0.01)})
\item \code{name} (character, layer name)
}
Any column not supplied falls back to \code{layer_dense()}'s own default.
Columns holding actual keras objects (regularizers, and
initializers if you don't just want a string) must be \emph{list-columns}
so each row can carry a different object or NULL, e.g.:
\preformatted{
tibble::tibble(
  units = c(64, 32, 1),
  activation = c("relu", "relu", "linear"),
  kernel_regularizer = list(keras3::regularizer_l2(0.001), NULL, NULL)
)
}}

\item{default_activation}{Activation function to be used if not otherwise
specified in \code{layer_df}.}
}
\value{
A function of a single argument \code{inputs} (a keras input
tensor) that returns a \code{keras3::keras_model()}.
}
\description{
Returns a \emph{function} \code{function(inputs) {...}} — not a model itself —
that builds a feed-forward stack of \code{layer_dense()} calls on top of
whatever \code{inputs} tensor it is later called with, and returns the
resulting \code{keras3::keras_model()}.
}
\details{
Layers can be specified either as a simple vector of unit counts, or
as a data frame / tibble giving full per-layer control over
activation, use_bias, initializers, regularizers, layer name, etc.
}
\examples{
\dontshow{if (reticulate::py_module_available(tensorflow)) withAutoprint(\{ # examplesIf}
mod_units <- build_dnn(units = c(16, 8, 1))
mod_df <- build_dnn(layer_df = data.frame(
  units = c(32, 16, 8, 1),
  activation = c("relu", "relu", "relu", "linear"),
  use_bias = c(FALSE, FALSE, FALSE, TRUE)
))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/persistence.R
\name{load_onam}
\alias{load_onam}
\title{Load a fitted onam model from disk}
\usage{
load_onam(dir)
}
\arguments{
\item{dir}{Directory as created by \link{save_onam}.}
}
\value{
Returns the restored object of class \code{onam}.
}
\description{
Load a fitted onam model from disk
}
\seealso{
\link{save_onam}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var_decomposition.R
\name{plot.var_decomp}
\alias{plot.var_decomp}
\title{Plot variance decomposition}
\usage{
\method{plot}{var_decomp}(x, ...)
}
\arguments{
\item{x}{Object of class \code{var_decomp} as returned by \link{decompose}}

\item{...}{further arguments, currently unused}
}
\value{
Returns a 'ggplot2' object showing the fraction of total
variance explained by each interaction order
}
\description{
Plot variance decomposition
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_helper.R
\name{install_conda_env}
\alias{install_conda_env}
\title{Set up conda environment for keras functionality}
\usage{
install_conda_env(
  envname = "r-keras",
  python_version = "python=3.10",
  overwrite = FALSE
)
}
\arguments{
\item{envname}{Name for the conda environment to be created.}

\item{python_version}{Python version to be installed in the conda environment.}

\item{overwrite}{Should an existing conda environment of name \code{envname} be
overwritten if present?}
}
\value{
No return value, called for side effects
}
\description{
Helper function to install Keras and packages necessary for package
functionality into a conda environment. Use this function if
\code{\link[keras3:install_keras]{keras3::install_keras()}} does not work, esp. on windows machines.
}
\seealso{
\code{\link[keras3:install_keras]{keras3::install_keras()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/persistence.R
\name{save_onam}
\alias{save_onam}
\title{Save a fitted onam model to disk}
\usage{
save_onam(object, dir, overwrite = FALSE)
}
\arguments{
\item{object}{Object of class \code{onam} as returned by \link{onam}.}

\item{dir}{Directory the model is saved to. Created if it does not exist.}

\item{overwrite}{Should an existing directory of the same name be
overwritten?}
}
\value{
No return value, called for side effects.
}
\description{
An \code{onam} object holds one or more fitted 'keras' submodels per ensemble
member (in \code{object$ensemble[[i]]$model_list}), which cannot be
serialized with \code{saveRDS()} alone. This function saves each submodel to
its own \code{.keras} file and the remaining orthogonalization metadata
(weights, \code{model_info}, data, predictions, ...) to a single \code{.rds} file,
all inside \code{dir}. Use \link{load_onam} to restore the object.
}
\seealso{
\link{load_onam}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_inter_effect}
\alias{plot_inter_effect}
\title{Plot Interaction Effect}
\usage{
plot_inter_effect(
  object,
  feature1,
  feature2,
  interpolate = FALSE,
  labs = NULL,
  custom_colors = "spectral",
  n_interpolate = 200,
  include_main = FALSE
)
}
\arguments{
\item{object}{Either model of class \code{onam} as returned from \link{onam} or
model evaluation outcome as returned from \link{predict.onam}}

\item{feature1, feature2}{Effects to be plotted.}

\item{interpolate}{If TRUE, values will be interpolated for a smooth plot.
If FALSE (default), only observations in the data will be plotted.}

\item{labs}{An optional named vector that can contain axis labels and the
name of the effect. Expected vector names are 'xlab', 'ylab' and 'effect'.}

\item{custom_colors}{color palette object for the interaction plot. Default
is "spectral", returning a color palette based on the spectral theme.}

\item{n_interpolate}{number of values per coordinate axis to interpolate.
Ignored if 'interpolate = FALSE'.}

\item{include_main}{If TRUE, main effects for features feature1 and feature2
will be added to the interaction term to give combined effect of main and
interaction effects. Default is FALSE.}
}
\value{
Returns a 'ggplot2' object of the specified effect interaction
}
\description{
Plot Interaction Effect
}
\examples{
\dontshow{if (reticulate::py_module_available(tensorflow)) withAutoprint(\{ # examplesIf}
# Basic example for a simple ONAM-model
# Create training data
n <- 1000
x1 <- runif(n, -2, 2)
x2 <- runif(n, -2, 2)
y <- sin(x1) + ifelse(x2 > 0, pweibull(x2, shape = 3),
  pweibull(-x2, shape = 0.5)) +
  x1 * x2
data_train <- cbind(x1, x2, y)
# Define model
model_formula <- y ~ mod1(x1) + mod1(x2) +
  mod1(x1, x2)
mod1 <- function(inputs) {
  outputs <- inputs \%>\%
    layer_dense(units = 16, activation = "relu") \%>\%
    layer_dense(units = 8, activation = "linear",
                use_bias = TRUE) \%>\%
    layer_dense(units = 1, activation = "linear",
                use_bias = TRUE)
  keras_model(inputs, outputs)
}
list_of_deep_models <- list(mod1 = mod1)
# Fit model
mod <- onam(model_formula, list_of_deep_models,
            data_train, n_ensemble = 1, epochs = 10)
plot_inter_effect(mod, "x1", "x2")
\dontshow{\}) # examplesIf}
}
